"""Wrapper classes for sampling from Bilby's prior module."""

from citation_compass import CiteClass

from lightcurvelynx.base_models import FunctionNode


class BilbyPriorNode(FunctionNode, CiteClass):
    """The base class for sampling from Bilby's prior module.

    Attributes
    ----------
    prior : bilby.prior.PriorDict
        The Bilby prior object to sample from.

    Parameters
    ----------
    prior : dict or bilby.prior.PriorDict
        A dictionary mapping the names of the parameters to their prior distributions.
    seed : int, optional
        The seed to use.

    Note
    ----
    The BilbyPriorNode's `compute()` function does not use the random number generator passed to it.
    Rather it uses Bilby's internal random number generator. You can set the seed for this using
    the `seed` argument when initializing the node or by calling the `set_seed` method. However,
    you cannot control the samples from this node via a simulation-wide random number generator.

    References
    ----------
    @article{bilby_paper,
        author = "Ashton, Gregory and others",
        title = "{BILBY: A user-friendly Bayesian inference library for gravitational-wave astronomy}",
        eprint = "1811.02042",
        archivePrefix = "arXiv",
        primaryClass = "astro-ph.IM",
        doi = "10.3847/1538-4365/ab06fc",
        journal = "Astrophys. J. Suppl.",
        volume = "241",
        number = "2",
        pages = "27",
        year = "2019"
    }
    """

    def __init__(self, prior, seed=None, **kwargs):
        if seed is not None:
            self.set_seed(seed)

        # Set the prior. If the prior is provided as a dictionary, convert it to a Bilby
        # PriorDict object.
        if isinstance(prior, dict):
            try:
                from bilby.core.prior import PriorDict
            except ImportError as err:  # pragma: no cover
                raise ImportError(
                    "Bilby package is not installed be default. To use the bilby priors, "
                    "please install it. For example, you can install it with `pip install bilby`."
                ) from err
            prior = PriorDict(prior)
        if len(prior) == 0:
            raise ValueError("The provided prior is empty.")
        self.prior = prior

        # Set the outputs to be the names of the parameters in the prior.
        outputs = [param for param in prior]
        super().__init__(self._non_func, outputs=outputs, **kwargs)

    def set_seed(self, new_seed):
        """Update the random number generator's seed to a given value.

        Parameters
        ----------
        new_seed : int
            The given seed
        """
        # Import bilby to set the seed.
        try:
            from bilby.core.utils import random as bibly_random
        except ImportError as err:  # pragma: no cover
            raise ImportError(
                "Bilby package is not installed be default. To use the bilby priors, "
                "please install it. For example, you can install it with `pip install bilby`."
            ) from err
        bibly_random.seed(new_seed)

    def compute(self, graph_state, rng_info=None, **kwargs):
        """Sample from the wrapped prior.

        The input arguments are taken from the current graph_state and the outputs
        are written to graph_state.

        Parameters
        ----------
        graph_state : GraphState
            An object mapping graph parameters to their values. This object is modified
            in place as it is sampled.
        rng_info : numpy.random._generator.Generator, optional
            This random number generator is not used by this node. Instead, Bilby's internal
            random number generator is used. You can set the seed for this using the `seed`
            argument when initializing the node or by calling the `set_seed()` method.
        **kwargs : dict, optional
            Additional function arguments.

        Returns
        -------
        results : any
            The result of the computation. This return value is provided so that testing
            functions can easily access the results.

        Raises
        ------
        ValueError is func attribute is None.
        """
        # Sample from the Bilby prior model and extract the parameters from the dictionary.
        param_dict = self.prior.sample(graph_state.num_samples)
        results = []
        for key in self.outputs:
            values = param_dict[key]
            if graph_state.num_samples == 1:
                values = values[0]
            graph_state.set(self.node_string, key, values)
            results.append(values)
        return results



import matplotlib

matplotlib.use("Agg")  # Suppress the plots

import numpy as np
import pytest
from lightcurvelynx.astro_utils.passbands import Passband, PassbandGroup
from lightcurvelynx.astro_utils.sed_basis_models import SEDBasisModel


def test_sed_basis_model() -> None:
    """Test that we can create and query a simple SEDBasisModel object."""
    # Create a simple SED basis model with three filters that have overlapping ranges.
    wavelengths = np.array([3000, 4000, 5000, 6000, 7000, 8000, 9000, 10000, 11000, 12000], dtype=float)
    filter_data = {
        "u": np.array([0.0, 0.5, 0.5, 0.5, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], dtype=float),
        "g": np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.8, 0.8, 0.8, 0.0, 0.0], dtype=float),
        "r": np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.6, 0.6, 0.6, 0.0], dtype=float),
    }
    sed_basis = SEDBasisModel(wavelengths, filter_data)
    assert len(sed_basis) == 3
    assert sed_basis.filters == ["u", "g", "r"]
    assert np.allclose(sed_basis.wavelengths, wavelengths)

    # Query the SED values for the bands at specific wavelengths.
    query_waves = np.array([3500, 4500, 5500, 6500, 7500, 8500, 9250])
    assert np.allclose(sed_basis.compute_sed("u", query_waves), [0.25, 0.5, 0.5, 0.25, 0.0, 0.0, 0.0])
    assert np.allclose(sed_basis.compute_sed("g", query_waves), [0.0, 0.0, 0.0, 0.0, 0.4, 0.8, 0.8])
    assert np.allclose(sed_basis.compute_sed("r", query_waves), [0.0, 0.0, 0.0, 0.0, 0.0, 0.3, 0.6])

    # We fall back to all internal wavelengths if none are provided.
    assert np.allclose(sed_basis.compute_sed("u"), filter_data["u"])
    assert np.allclose(sed_basis.compute_sed("g"), filter_data["g"])
    assert np.allclose(sed_basis.compute_sed("r"), filter_data["r"])

    # We can compute the flux densities over time.
    flux_density = sed_basis.compute_sed_from_bandfluxes(
        ["r", "g", "g", "r"], bandfluxes=[2.0, 1.0, 1.5, 2.5]
    )
    expected = np.array(
        [
            sed_basis.compute_sed("r") * 2.0,
            sed_basis.compute_sed("g") * 1.0,
            sed_basis.compute_sed("g") * 1.5,
            sed_basis.compute_sed("r") * 2.5,
        ]
    )
    assert np.allclose(flux_density, expected)

    # SED creation fails if the lengths do not match.
    with pytest.raises(ValueError):
        _ = SEDBasisModel(np.array([300, 400, 500, 600, 700]), filter_data)

    # Test plot doesn't crash
    sed_basis.plot()


def test_create_from_passbands():
    """Test that we can create a SEDBasisModel from a PassbandGroup."""
    a_band = Passband(np.array([[4000, 0.5], [5000, 0.5], [6000, 0.5]]), "LSST", "u")
    b_band = Passband(np.array([[8000, 0.8], [9000, 0.8], [10000, 0.8]]), "LSST", "g")
    c_band = Passband(np.array([[9000, 0.6], [10000, 0.6], [11000, 0.6]]), "LSST", "r")
    passband_group = PassbandGroup(given_passbands=[a_band, b_band, c_band])
    sed_basis = SEDBasisModel.from_box_approximation(passband_group)

    assert np.allclose(sed_basis.wavelengths, passband_group.waves)
    assert sed_basis.filters == ["u", "g", "r"]

    # Check that no two SED basis functions overlap.
    for f1 in sed_basis.filters:
        for f2 in sed_basis.filters:
            if f1 != f2:
                assert np.count_nonzero(sed_basis.sed_values[f1] * sed_basis.sed_values[f2]) == 0

    # Check that they are normalized correctly.
    for filter_name in sed_basis.filters:
        sed = sed_basis.sed_values[filter_name]
        flux_density = np.array([sed])
        bandflux = passband_group.fluxes_to_bandflux(flux_density, filter_name)
        assert np.isclose(bandflux, 1.0)



import logging

import numpy as np

from lightcurvelynx.astro_utils.noise_model import poisson_bandflux_std
from lightcurvelynx.obstable.obs_table import ObsTable
from lightcurvelynx.obstable.obs_table_params import ParamDeriver

logger = logging.getLogger(__name__)


class FakeObsTable(ObsTable):
    """A subclass for a (simplified) fake survey. The user must provide a constant
    flux error to use or enough information to compute the poisson_bandflux_std noise model.

    The class uses a flexible deriver to try to compute any missing parameters needed from
    what is provided.

    Defaults are set for other parameters (e.g. exptime, nexposure, read_noise, dark_current), which
    the user can override with keyword arguments to the constructor.

    Parameters
    ----------
    table : dict or pandas.core.frame.DataFrame
        The table with all the ObsTable information.  Must have columns
        "time", "ra", "dec", and "filter".
    colmap : dict, optional
        A mapping of standard column names to their names in the input table.
    zp_per_band : dict, optional
        A dictionary mapping filter names to their instrumental zero points (flux in nJy
        corresponding to 1 electron per exposure). The filters provided must match those
        in the table. This is required if the table does not have a zero point column.
    const_flux_error : float or dict, optional
        If provided, use this constant flux error (in nJy) for all observations (overriding
        the normal noise compuation). A value of 0.0 will produce a noise-free simulation.
        If a dictionary is provided, it should map filter names to constant flux errors per-band.
        This setting should primarily be used for testing purposes.
    radius : float, optional
        The angular radius of the field of view of the observations in degrees (default=None).
    saturation_mags : dict, optional
        A dictionary mapping filter names to their saturation thresholds in magnitudes. The filters
        provided must match those in the table. If not provided, saturation effects will not be applied.
    noise_strategy : str, optional
        The name of the strategy to use to derive any missing table parameters needed to compute the noise.
        This is used if the table does not provide all the necessary parameters for the given noise model.
        Should be one of:
        - "given_only" : Use only the parameters already provided in the table and survey values.
        - "five_sigma_depth": Derive approximate noise from only the 5-sigma depth values if available
          (no survey values like PSF, sky background, etc. are used).
        - "exhaustive": Try all available derivation methods to fill in missing parameters.
        Default is "given_only" which does not attempt any derivation.
    **kwargs : dict
        Additional keyword arguments to pass to the ObsTable constructor. This includes overrides
        for survey parameters such as:
        - survey_name: The name of the survey (default="FAKE_SURVEY").
    """

    # Default survey values.
    _default_survey_values = {
        "dark_current": 0,
        "exptime": 30,  # seconds
        "fwhm_px": None,  # pixels
        "nexposure": 1,  # exposures
        "radius": None,  # degrees
        "read_noise": 0,  # electrons
        "sky_bg_electrons": None,  # electrons / pixel^2
        "survey_name": "FAKE_SURVEY",
    }

    def __init__(
        self,
        table,
        *,
        colmap=None,
        const_flux_error=None,
        noise_strategy="given_only",
        **kwargs,
    ):
        # Pass along all the survey parameters to the parent class.
        super().__init__(
            table,
            colmap=colmap,
            **kwargs,
        )

        # Derive any missing parameters needed for the flux error computation. We always create
        # a new ParamDeriver instance here, because they are stateful.
        if noise_strategy == "given_only":
            param_deriver = "NoopParamDeriver"
        elif noise_strategy == "five_sigma_depth":
            param_deriver = "FiveSigmaDepthParamDeriver"
        elif noise_strategy == "exhaustive":
            param_deriver = "FullParamDeriver"
        else:
            raise ValueError(
                f"Invalid noise_strategy '{noise_strategy}'. "
                "Should be one of: 'given_only', 'five_sigma_depth', 'exhaustive'."
            )
        param_deriver_obj = ParamDeriver.create_deriver(param_deriver)
        param_deriver_obj.derive_parameters(self)

        # If a constant flux error is provided, validate it.
        if const_flux_error is not None:
            # Convert a constant into a per-band dictionary.
            if isinstance(const_flux_error, int | float):
                const_flux_error = {fil: const_flux_error for fil in self.filters}

            # Check that every filter occurs in the dictionary.
            for fil in self.filters:
                if fil not in const_flux_error:
                    raise ValueError(
                        "`const_flux_error` must include all the filters in the table. Missing '{fil}'."
                    )

            # Translate the constant flux errors into a table column.
            bandflux_errors = np.array([const_flux_error[fil] for fil in self._table["filter"]])
            if np.any(bandflux_errors < 0):
                raise ValueError("Constant flux errors must be non-negative.")
            self._table["bandflux_error"] = bandflux_errors

        # Validate that we have enough information to compute flux errors.
        if "zp" not in self._table and "bandflux_error" not in self._table:
            column_names = self._table.columns.tolist()
            param_names = list(self.survey_values.keys())
            raise ValueError(
                f"Insufficient information to compute flux errors or zeropoints using {param_deriver}. "
                f"Table columns: {column_names}. Survey parameters: {param_names}."
            )

    def bandflux_error_point_source(self, bandflux, index):
        """Compute observational bandflux error for a point source

        Parameters
        ----------
        bandflux : array_like of float
            Band bandflux of the point source in nJy.
        index : array_like of int
            The index of the observation in the OpSim table.

        Returns
        -------
        flux_err : array_like of float
            Simulated bandflux noise in nJy.
        """
        # If we have a bandflux_error in the table (including from constant flux error), use that.
        if "bandflux_error" in self._table:
            return self.get_value_per_row("bandflux_error", indices=index)

        # Otherwise compute the flux error using the poisson_bandflux_std noise model.
        return poisson_bandflux_std(
            bandflux,
            total_exposure_time=self.get_value_per_row("exptime", indices=index),
            exposure_count=self.get_value_per_row("nexposure", indices=index),
            psf_footprint=self.get_value_per_row("psf_footprint", indices=index),
            sky=self.get_value_per_row("sky_bg_electrons", indices=index),
            zp=self.get_value_per_row("zp", indices=index),
            readout_noise=self.safe_get_survey_value("read_noise"),
            dark_current=self.safe_get_survey_value("dark_current"),
        )



"""Utility functions for handling redshift and cosmological calculations."""

from functools import partial

import astropy.cosmology.units as cu
from astropy import units as u

from lightcurvelynx.base_models import FunctionNode


def obs_to_rest_times_waves(observer_frame_times, observer_frame_wavelengths, redshift, t0):
    """Calculate the rest frame times and wavelengths needed to give user the observer frame times
    and wavelengths (given the redshift).

    Parameters
    ----------
    observer_frame_times : numpy.ndarray
        The times at which the observation is made.
    observer_frame_wavelengths : numpy.ndarray
        The wavelengths at which the observation is made (in Angstroms).
    redshift : float
        The redshift of the object.
    t0 : float
        The reference epoch for the given object.

    Returns
    -------
    tuple of (numpy.ndarray, numpy.ndarray)
        The rest frame times and wavelengths needed to generate the rest frame flux densities,
        which will later be redshifted  back to observer frame flux densities at the observer frame
        times and wavelengths.
    """
    observed_times_rel_to_t0 = observer_frame_times - t0
    rest_frame_times_rel_to_t0 = observed_times_rel_to_t0 / (1 + redshift)
    rest_frame_times = rest_frame_times_rel_to_t0 + t0
    rest_frame_wavelengths = observer_frame_wavelengths / (1 + redshift)
    return (rest_frame_times, rest_frame_wavelengths)


def rest_to_obs_flux(flux_density, redshift):
    """Convert rest-frame flux to obs-frame flux.
    The (1+redshift) factor is applied to preserve bolometric flux.
    The rest-frame flux is defined as F_nu = L_nu / 4*pi*D_L**2,
    where D_L is the luminosity distance.

    Parameters
    ----------
    flux_density : numpy.ndarray
        A length T X N matrix of flux density values (in nJy).
    redshift : float
        The redshift of the object associated with given flux densities.

    Returns
    -------
    flux_density : numpy.ndarray
        The observer frame flux (in nJy).
    """
    # Note that the multiplication by (1+z) is due to the fact we are working in f_nu units,
    # instead of f_lambda units and may be unintuitive for users who are used to working in f_lambda
    # units. This factor can be derived by equaling the integrated flux in f_nu unit before and after
    # redshift is applied.
    return flux_density * (1 + redshift)


def redshift_to_distance(redshift, cosmology):
    """Compute a source's luminosity distance given its redshift and a
    specified cosmology using astropy's redshift_distance().

    Parameters
    ----------
    redshift : float
        The redshift value.
    cosmology : astropy.cosmology
        The cosmology specification.

    Returns
    -------
    distance : float
        The luminosity distance (in pc)
    """
    z = redshift * cu.redshift
    distance = z.to(u.pc, cu.redshift_distance(cosmology, kind="luminosity"))
    return distance.value


class RedshiftDistFunc(FunctionNode):
    """A wrapper class for the redshift_to_distance() function.

    Attributes
    ----------
    cosmology : astropy.cosmology
        The cosmology specification.

    Parameters
    ----------
    redshift : function or constant
        The function or constant providing the redshift value.
    cosmology : astropy.cosmology
        The cosmology specification.
    **kwargs : dict, optional
        Any additional keyword arguments.
    """

    def __init__(self, redshift, cosmology, **kwargs):
        # Create a partial function with the cosmology fixed. We do this
        # so that cosmology is not added as a node parameter.
        func = partial(redshift_to_distance, cosmology=cosmology)
        func.__name__ = "redshift_to_distance"
        func.__doc__ = redshift_to_distance.__doc__

        # Call the super class's constructor with the needed information.
        super().__init__(
            func=func,
            redshift=redshift,
            **kwargs,
        )



import numpy as np
import pytest
from lightcurvelynx.astro_utils.dustmap import ConstantHemisphereDustMap, DustmapWrapper
from lightcurvelynx.effects.extinction import ExtinctionEffect
from lightcurvelynx.math_nodes.given_sampler import GivenValueList
from lightcurvelynx.models.basic_models import ConstantSEDModel


def test_list_extinction_models():
    """List the available extinction models."""
    model_names = ExtinctionEffect.list_extinction_models()
    assert len(model_names) > 10
    assert "G23" in model_names
    assert "CCM89" in model_names


def test_load_extinction_model():
    """Load an extinction model by string."""
    g23_model = ExtinctionEffect.load_extinction_model("G23", Rv=3.1)
    assert g23_model is not None
    assert hasattr(g23_model, "extinguish")

    # We fail if we try to load a model that does not exist.
    with pytest.raises(KeyError):
        ExtinctionEffect.load_extinction_model("InvalidModel")

    # We can manually load the g23_model into an ExtinctionEffect node.
    dust_effect = ExtinctionEffect(g23_model, ebv=0.1, frame="rest")

    # We can apply the extinction effect to a set of fluxes.
    fluxes = np.full((10, 3), 1.0)
    wavelengths = np.array([7000.0, 5200.0, 4800.0])
    new_fluxes = dust_effect.apply(fluxes, wavelengths=wavelengths, ebv=0.1)
    assert new_fluxes.shape == (10, 3)
    assert np.all(new_fluxes < fluxes)

    # We fail if we are missing a required parameter.
    with pytest.raises(ValueError):
        _ = dust_effect.apply(fluxes, wavelengths=wavelengths)
    with pytest.raises(ValueError):
        _ = dust_effect.apply(fluxes, ebv=0.1)


def test_set_frame():
    """Test that correct frame is set"""
    ext = ExtinctionEffect("G23", ebv=0.1, frame="observer")
    assert ext.rest_frame is False

    with pytest.raises(ValueError):
        ExtinctionEffect("G23", ebv=0.1, frame="InvalidFrame")


def test_constant_dust_extinction():
    """Test that we can create and sample a ExtinctionEffect object."""
    # Use given ebv values. Usually these would be computed from a dustmap,
    # based on (RA, dec).
    ebv_node = GivenValueList([0.1, 0.2, 0.3, 0.4, 0.5])
    dust_effect = ExtinctionEffect("CCM89", ebv=ebv_node, Rv=3.1, frame="rest")
    assert dust_effect.extinction_model is not None
    assert hasattr(dust_effect.extinction_model, "extinguish")

    model = ConstantSEDModel(
        brightness=100.0,
        ra=0.0,
        dec=40.0,
        redshift=0.0,
    )
    model.add_effect(dust_effect)

    times = np.array([1.0, 2.0, 3.0, 4.0, 5.0])
    wavelengths = np.array([7000.0, 5200.0, 4800.0])  # Red, green, blue
    states = model.sample_parameters(num_samples=3)
    fluxes = model.evaluate_sed(times, wavelengths, states)

    assert fluxes.shape == (3, 5, 3)
    assert np.all(fluxes < 100.0)
    assert np.all(fluxes[0, :, :] > fluxes[1, :, :])
    assert np.all(fluxes[1, :, :] > fluxes[2, :, :])


def test_dustmap_chain():
    """Test that we can chain the dustmap computation and extinction effect."""
    model = ConstantSEDModel(
        brightness=100.0,
        ra=GivenValueList([45.0, 45.0, 45.0]),
        dec=GivenValueList([20.0, -20.0, 10.0]),
        redshift=0.0,
    )

    # Create a constant dust map for testing.
    dust_map = ConstantHemisphereDustMap(north_ebv=0.8, south_ebv=0.5)
    dust_map_node = DustmapWrapper(dust_map, ra=model.ra, dec=model.dec)

    # Create an extinction effect using the EBVs from that dust map.
    ext_effect = ExtinctionEffect(extinction_model="CCM89", ebv=dust_map_node, Rv=3.1, frame="rest")
    model.add_effect(ext_effect)

    # Sample the model.
    times = np.array([1.0, 2.0, 3.0, 4.0, 5.0])
    wavelengths = np.array([7000.0, 5200.0])
    states = model.sample_parameters(num_samples=3)
    fluxes = model.evaluate_sed(times, wavelengths, states)

    assert fluxes.shape == (3, 5, 2)
    assert np.all(fluxes < 100.0)
    assert np.allclose(fluxes[0, :, :], fluxes[2, :, :])
    assert np.all(fluxes[1, :, :] > fluxes[0, :, :])



import tempfile
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
from pathlib import Path

import numpy as np
import pandas as pd
import pytest
from lightcurvelynx.astro_utils.mag_flux import mag2flux
from lightcurvelynx.astro_utils.passbands import Passband, PassbandGroup
from lightcurvelynx.graph_state import GraphState
from lightcurvelynx.math_nodes.given_sampler import GivenValueList
from lightcurvelynx.math_nodes.np_random import NumpyRandomFunc
from lightcurvelynx.models.basic_models import ConstantSEDModel, StepModel
from lightcurvelynx.models.static_sed_model import StaticBandfluxModel
from lightcurvelynx.obstable.fake_obs_table import FakeObsTable
from lightcurvelynx.obstable.opsim import OpSim
from lightcurvelynx.obstable.ztf_obstable import ZTFObsTable, create_random_ztf_obs_data
from lightcurvelynx.simulate import (
    SimulationInfo,
    compute_noise_free_lightcurves,
    compute_single_noise_free_lightcurve,
    get_time_windows,
    simulate_lightcurves,
)
from nested_pandas import read_parquet


def test_get_time_windows():
    """Test the get_time_windows function with various inputs."""
    assert get_time_windows(None, None) == (None, None)
    assert get_time_windows(0.0, None) == (None, None)
    assert get_time_windows(None, (1.0, 2.0)) == (None, None)

    result = get_time_windows(0.0, (-1.0, 2.0))
    assert np.array_equal(result[0], np.array([-1.0]))
    assert np.array_equal(result[1], np.array([2.0]))

    result = get_time_windows(1.0, (None, 2.0))
    assert result[0] is None
    assert np.array_equal(result[1], np.array([3.0]))

    result = get_time_windows(-10.0, (-1.0, None))
    assert np.array_equal(result[0], np.array([-11.0]))
    assert result[1] is None

    result = get_time_windows(np.array([0.0, 1.0, 2.0]), (-1.0, 2.0))
    assert np.array_equal(result[0], np.array([-1.0, 0.0, 1.0]))
    assert np.array_equal(result[1], np.array([2.0, 3.0, 4.0]))

    with pytest.raises(ValueError):
        get_time_windows(0.0, (1.0, 2.0, 3.0))


def test_simulation_info():
    """Test that we can create and split SimulationInfo objects."""
    model = ConstantSEDModel(brightness=100.0, t0=0.0, ra=0.0, dec=0.0, redshift=0.0, node_label="source")

    # Create a completely fake passband group and obstable for testing.
    pb_group = PassbandGroup(
        [
            Passband(np.array([[100, 0.5], [200, 0.75], [300, 0.25]]), "my_survey", "a"),
            Passband(np.array([[250, 0.25], [300, 0.5], [350, 0.75]]), "my_survey", "b"),
        ]
    )

    values = {
        "time": np.array([0.0, 1.0, 2.0, 3.0, 4.0]),
        "ra": np.array([15.0, 30.0, 15.0, 0.0, 60.0]),
        "dec": np.array([-10.0, -5.0, 0.0, 5.0, 10.0]),
        "filter": np.array(["r", "g", "r", "i", "g"]),
    }
    zp_per_band = {"g": 26.0, "r": 27.0, "i": 28.0}
    ops_data = FakeObsTable(
        values,
        noise_strategy="exhaustive",
        zp_per_band=zp_per_band,
        fwhm_px=2.0,
        sky_bg_electrons=100.0,
    )

    sim_info = SimulationInfo(
        model=model,
        num_samples=100,
        obstable=ops_data,
        passbands=pb_group,
        time_window_offset=(-5.0, 10.0),  # A keyword argument to test
        rng=np.random.default_rng(12345),
    )
    assert sim_info.num_samples == 100
    assert sim_info.time_window_offset == (-5.0, 10.0)

    # Test splitting into batches.
    batches = sim_info.split(num_batches=3)
    assert len(batches) == 3
    assert batches[0].num_samples == 34
    assert batches[1].num_samples == 34
    assert batches[2].num_samples == 32

    # We propagate all the keyword parameters.
    assert batches[0].time_window_offset == (-5.0, 10.0)
    assert batches[1].time_window_offset == (-5.0, 10.0)
    assert batches[2].time_window_offset == (-5.0, 10.0)

    # They have different RNGs with different sampling states.
    assert batches[0].rng is not batches[1].rng
    assert batches[0].rng is not batches[2].rng
    assert batches[1].rng is not batches[2].rng
    sample0 = batches[0].rng.integers(0, 100000)
    sample1 = batches[1].rng.integers(0, 100000)
    sample2 = batches[2].rng.integers(0, 100000)
    assert sample0 != sample1
    assert sample0 != sample2
    assert sample1 != sample2

    with pytest.raises(ValueError):
        # Neither num_samples nor num_batches is specified.
        sim_info.split()
    with pytest.raises(ValueError):
        # num_samples and batch_size are both specified.
        sim_info.split(batch_size=10, num_batches=2)
    with pytest.raises(ValueError):
        # Negative num_batches.
        sim_info.split(num_batches=-1)
    with pytest.raises(ValueError):
        # Negative batch_size.
        sim_info.split(batch_size=-10)

    # Fail with a negative number of samples.
    with pytest.raises(ValueError):
        _ = SimulationInfo(
            model=model,
            num_samples=-10,
            obstable=ops_data,
            passbands=pb_group,
            time_window_offset=(-5.0, 10.0),  # A keyword argument to test
            rng=np.random.default_rng(12345),
        )


def test_simulate_lightcurves(test_data_dir):
    """Test an end to end run of simulating the light curves."""
    # Load the OpSim data.
    opsim_db = OpSim.from_db(test_data_dir / "opsim_small.db")

    # Load the passband data for the griz filters only.
    passband_group = PassbandGroup.from_preset(
        preset="LSST",
        table_dir=test_data_dir / "passbands",
        filters=["g", "r", "i", "z"],
    )

    # Create a constant SED model with known brightnesses and RA, dec
    # values that match the opsim.
    given_brightness = [1000.0, 2000.0, 5000.0, 1000.0, 100.0]
    source = ConstantSEDModel(
        brightness=GivenValueList(given_brightness),
        t0=0.0,
        ra=GivenValueList(opsim_db["ra"].values[0:5]),
        dec=GivenValueList(opsim_db["dec"].values[0:5]),
        redshift=0.0,
        node_label="source",
    )

    results = simulate_lightcurves(
        source,
        5,
        opsim_db,
        passband_group,
        obstable_save_cols=["observationId", "zp_nJy"],
        param_cols=["source.brightness"],
    )
    assert len(results) == 5
    assert np.all(results["nobs"].values >= 1)
    assert np.allclose(results["ra"].values, opsim_db["ra"].values[0:5])
    assert np.allclose(results["dec"].values, opsim_db["dec"].values[0:5])
    assert np.allclose(results["z"].values, 0.0)
    assert np.allclose(results["t0"].values, 0.0)

    for idx in range(5):
        num_obs = results["nobs"][idx]
        assert len(results.loc[idx]["lightcurve"]["flux"]) == num_obs

        # Check that we pulled the metadata from the opsim.
        assert len(results.loc[idx]["lightcurve"]["observationId"]) == num_obs
        assert len(np.unique(results.loc[idx]["lightcurve"]["observationId"])) == num_obs
        assert np.all(results.loc[idx]["lightcurve"]["observationId"] >= 0)
        assert len(results.loc[idx]["lightcurve"]["zp_nJy"]) == num_obs

        # Check that we extract one of the parameters.
        assert results["source_brightness"][idx] == given_brightness[idx]

    # Check that we saved and can reassemble the GraphStates
    assert "params" in results
    state = GraphState.from_list(results["params"].values)
    assert state.num_samples == 5
    assert np.allclose(state["source.ra"], opsim_db["ra"].values[0:5])
    assert np.allclose(state["source.dec"], opsim_db["dec"].values[0:5])

    # Check that we can extract a single GraphState from the results.
    single_state = GraphState.from_dict(results["params"][2])
    assert single_state.num_samples == 1
    assert single_state["source.ra"] == opsim_db["ra"].values[2]
    assert single_state["source.dec"] == opsim_db["dec"].values[2]

    # Check that we fail if we try to save a parameter column that doesn't exist.
    # And that the error gives the existing options.
    source2 = ConstantSEDModel(
        brightness=10.0,
        t0=0.0,
        ra=1.0,
        dec=-1.0,
        redshift=0.0,
        node_label="source2",
    )
    with pytest.raises(KeyError) as excinfo:
        _ = simulate_lightcurves(
            source2,
            1,
            opsim_db,
            passband_group,
            param_cols=["source2.unknown_parameter"],
        )
    assert "Available parameters are:" in str(excinfo.value)
    assert "source2.ra" in str(excinfo.value)
    assert "source2.dec" in str(excinfo.value)


def test_simulate_lightcurves_to_file(test_data_dir):
    """Test an end to end run of simulating the light curves, saving them to a file."""
    # Load the OpSim data.
    opsim_db = OpSim.from_db(test_data_dir / "opsim_small.db")

    # Load the passband data for the griz filters only.
    passband_group = PassbandGroup.from_preset(
        preset="LSST",
        table_dir=test_data_dir / "passbands",
        filters=["g", "r", "i", "z"],
    )

    # Create a constant SED model with known brightnesses and RA, dec
    # values that match the opsim.
    given_brightness = [1000.0, 2000.0, 5000.0, 1000.0, 100.0]
    source = ConstantSEDModel(
        brightness=GivenValueList(given_brightness),
        t0=0.0,
        ra=GivenValueList(opsim_db["ra"].values[0:5]),
        dec=GivenValueList(opsim_db["dec"].values[0:5]),
        redshift=0.0,
        node_label="source",
    )

    with tempfile.TemporaryDirectory() as tmpdir:
        file_path = Path(tmpdir) / "test_simulate_lightcurves.parquet"
        assert not file_path.exists()

        results = simulate_lightcurves(
            source,
            5,
            opsim_db,
            passband_group,
            obstable_save_cols=["observationId", "zp_nJy"],
            param_cols=["source.brightness"],
            output_file_path=file_path,
        )
        assert str(results) == str(file_path)
        assert file_path.exists()

        results_df = read_parquet(file_path)
        assert len(results_df) == 5
        assert np.all(results_df["nobs"].to_numpy() >= 1)
        assert np.allclose(results_df["ra"].to_numpy(), opsim_db["ra"].values[0:5])
        assert np.allclose(results_df["dec"].to_numpy(), opsim_db["dec"].values[0:5])
        assert np.allclose(results_df["z"].to_numpy(), 0.0)
        assert np.allclose(results_df["t0"].to_numpy(), 0.0)


def test_simulate_bandfluxes(test_data_dir):
    """Test an end to end run of simulating a bandflux model."""
    # Create a toy observation table with two pointings.
    num_obs = 6
    obsdata = {
        "time": [0.0, 1.0, 2.0, 3.0, 4.0, 5.0],
        "ra": [0.0, 0.0, 180.0, 0.0, 180.0, 0.0],
        "dec": [10.0, 10.0, -10.0, 10.0, -10.0, 10.0],
        "filter": ["g", "r", "g", "r", "z", "z"],
        "zp": [1.0] * num_obs,
        "seeing": [1.12] * num_obs,
        "skybrightness": [20.0] * num_obs,
        "exptime": [29.2] * num_obs,
        "nexposure": [2] * num_obs,
    }
    obstable = OpSim(obsdata)

    # Load the passband data for the griz filters only.
    passband_group = PassbandGroup.from_preset(
        preset="LSST",
        table_dir=test_data_dir / "passbands",
        filters=["g", "r", "i", "z"],
    )

    # Create a static bandflux model and simulate 2 runs.
    model = StaticBandfluxModel({"g": 1.0, "i": 2.0, "r": 3.0, "y": 4.0, "z": 5.0}, ra=0.0, dec=10.0)
    results = simulate_lightcurves(model, 2, obstable, passband_group)
    assert len(results) == 2
    assert np.allclose(results["lightcurve"][0]["flux_perfect"], [1.0, 3.0, 3.0, 5.0])
    assert np.allclose(results["lightcurve"][1]["flux_perfect"], [1.0, 3.0, 3.0, 5.0])


def test_simulate_parallel_threads(test_data_dir):
    """Test an end to end run of simulating the light curves parallel with threads."""
    # Load the OpSim data.
    opsim_db = OpSim.from_db(test_data_dir / "opsim_small.db")

    # Load the passband data for the griz filters only.
    passband_group = PassbandGroup.from_preset(
        preset="LSST",
        table_dir=test_data_dir / "passbands",
        filters=["g", "r", "i", "z"],
    )

    # Create a constant SED model with known brightnesses and RA, dec
    # values that match the opsim.
    ra0 = opsim_db["ra"].values[0]
    dec0 = opsim_db["dec"].values[0]
    source = ConstantSEDModel(
        brightness=NumpyRandomFunc("uniform", low=100.0, high=500.0),
        t0=0.0,
        ra=NumpyRandomFunc("uniform", low=ra0 - 0.5, high=ra0 + 0.5),
        dec=NumpyRandomFunc("uniform", low=dec0 - 0.5, high=dec0 + 0.5),
        redshift=0.0,
        node_label="source",
    )

    with ThreadPoolExecutor(max_workers=2) as executor:
        results = simulate_lightcurves(
            source,
            100,
            opsim_db,
            passband_group,
            obstable_save_cols=["observationId", "zp_nJy"],
            param_cols=["source.brightness"],
            executor=executor,
            batch_size=10,
        )
    assert len(results) == 100
    assert np.all(results["nobs"].values >= 1)
    assert np.all(results["ra"].values >= ra0 - 0.5)
    assert np.all(results["ra"].values <= ra0 + 0.5)
    assert np.all(results["dec"].values >= dec0 - 0.5)
    assert np.all(results["dec"].values <= dec0 + 0.5)

    for idx in range(100):
        num_obs = results["nobs"][idx]
        assert num_obs >= 1
        assert len(results["lightcurve"][idx]["flux"]) == num_obs


def test_simulate_parallel_processes(test_data_dir):
    """Test an end to end run of simulating the light curves paralle with processes."""
    # Load the OpSim data.
    opsim_db = OpSim.from_db(test_data_dir / "opsim_small.db")

    # Load the passband data for the griz filters only.
    passband_group = PassbandGroup.from_preset(
        preset="LSST",
        table_dir=test_data_dir / "passbands",
        filters=["g", "r", "i", "z"],
    )

    # Create a constant SED model with known brightnesses and RA, dec
    # values that match the opsim.
    ra0 = opsim_db["ra"].values[0]
    dec0 = opsim_db["dec"].values[0]
    source = ConstantSEDModel(
        brightness=NumpyRandomFunc("uniform", low=100.0, high=500.0),
        t0=0.0,
        ra=NumpyRandomFunc("uniform", low=ra0 - 0.5, high=ra0 + 0.5),
        dec=NumpyRandomFunc("uniform", low=dec0 - 0.5, high=dec0 + 0.5),
        redshift=0.0,
        node_label="source",
    )

    with ProcessPoolExecutor(max_workers=2) as executor:
        results = simulate_lightcurves(
            source,
            100,
            opsim_db,
            passband_group,
            obstable_save_cols=["observationId", "zp_nJy"],
            param_cols=["source.brightness"],
            executor=executor,
            batch_size=10,
        )
    assert len(results) == 100
    assert np.all(results["nobs"].values >= 1)
    assert np.all(results["ra"].values >= ra0 - 0.5)
    assert np.all(results["ra"].values <= ra0 + 0.5)
    assert np.all(results["dec"].values >= dec0 - 0.5)
    assert np.all(results["dec"].values <= dec0 + 0.5)

    for idx in range(100):
        num_obs = results["nobs"][idx]
        assert num_obs >= 1
        assert len(results["lightcurve"][idx]["flux"]) == num_obs

    # We can use the default (ProcessPoolExecutor) by giving a number of jobs.
    results2 = simulate_lightcurves(
        source,
        100,
        opsim_db,
        passband_group,
        obstable_save_cols=["observationId", "zp_nJy"],
        param_cols=["source.brightness"],
        batch_size=10,
        num_jobs=2,
    )
    assert len(results2) == 100

    # We can write the results to files.
    with tempfile.TemporaryDirectory() as tmpdir:
        base_path = Path(tmpdir) / "test_df.parquet"
        ind_paths = []
        for i in range(4):
            curr_path = Path(tmpdir) / f"test_df_part{i}.parquet"
            assert not curr_path.exists()
            ind_paths.append(curr_path)

        results3 = simulate_lightcurves(
            source,
            200,
            opsim_db,
            passband_group,
            obstable_save_cols=["observationId", "zp_nJy"],
            param_cols=["source.brightness"],
            output_file_path=base_path,
            batch_size=50,
            num_jobs=4,
        )
        assert len(results3) == 4
        for i in range(4):
            assert str(results3[i]) == str(ind_paths[i])
            assert ind_paths[i].exists()

            results_df = read_parquet(ind_paths[i])
            assert len(results_df) == 50


def test_simulate_single_lightcurve(test_data_dir):
    """Test an end to end run of simulating a single light curve."""
    # Load the OpSim data.
    opsim_db = OpSim.from_db(test_data_dir / "opsim_small.db")

    # Load the passband data for the griz filters only.
    passband_group = PassbandGroup.from_preset(
        preset="LSST",
        table_dir=test_data_dir / "passbands",
        filters=["g", "r", "i", "z"],
    )

    # Create a constant SED model with known brightnesses and RA, dec
    # values that match the opsim.
    given_brightness = [1000.0, 2000.0, 5000.0, 1000.0, 100.0]
    source = ConstantSEDModel(
        brightness=GivenValueList(given_brightness),
        t0=0.0,
        ra=GivenValueList(opsim_db["ra"].values[0:5]),
        dec=GivenValueList(opsim_db["dec"].values[0:5]),
        redshift=0.0,
        node_label="source",
    )

    results = simulate_lightcurves(
        source,
        1,
        opsim_db,
        passband_group,
        obstable_save_cols=["observationId", "zp_nJy"],
        param_cols=["source.brightness"],
    )
    assert len(results) == 1

    # Check that we saved and can reassemble the GraphStates
    assert "params" in results
    state = GraphState.from_list(results["params"].values)
    assert state.num_samples == 1
    assert state["source.ra"] == opsim_db["ra"].values[0]
    assert state["source.dec"] == opsim_db["dec"].values[0]


def test_simulate_with_time_window(test_data_dir):
    """Test an end to end run of simulating with a limited time window."""
    # Create a toy OpSim database with two pointings over a series of tiems.
    # Create a fake opsim data frame with just time, RA, and dec.
    values = {
        "observationStartMJD": np.arange(50.0),
        "fieldRA": np.array([15.0 if i % 2 == 0 else 180.0 for i in range(50)]),
        "fieldDec": np.array([10.0 if i % 2 == 0 else -10.0 for i in range(50)]),
        "filter": np.full(50, "g"),
        # We add the remaining values so the OpSim can compute noise, but they are
        # arbitrary and not tested in this test.
        "zp_nJy": np.ones(50),
        "seeingFwhmEff": np.ones(50) * 0.7,
        "visitExposureTime": np.ones(50) * 30.0,
        "numExposures": np.ones(50) * 1,
        "skyBrightness": np.full(50, 20.0),
    }
    opsim_db = OpSim(values)

    # Load the passband data for the griz filters only.
    passband_group = PassbandGroup.from_preset(
        preset="LSST",
        table_dir=test_data_dir / "passbands",
        filters=["g", "r", "i", "z"],
    )

    # Create a constant SED model with known brightnesses and RA, dec
    # values that match the opsim.
    source = ConstantSEDModel(
        brightness=1000.0,
        t0=GivenValueList([20.0, 15.0]),
        ra=15.0,
        dec=10.0,
        redshift=0.0,
        node_label="source",
    )

    results = simulate_lightcurves(
        source,
        2,
        opsim_db,
        passband_group,
        time_window_offset=(-5.0, 10.0),
    )
    assert len(results) == 2

    # We should simulate the observations that are only within the time window, (15.0, 30.0) for the
    # first samples and (10.0, 25.0) for the second sample, and at the matching RA/Dec (the even indices).
    assert np.array_equal(
        results["lightcurve"][0]["mjd"],
        np.array([16.0, 18.0, 20.0, 22.0, 24.0, 26.0, 28.0, 30.0]),
    )
    assert np.array_equal(
        results["lightcurve"][1]["mjd"],
        np.array([10.0, 12.0, 14.0, 16.0, 18.0, 20.0, 22.0, 24.0]),
    )


def test_simulate_multiple_surveys(test_data_dir):
    """Test an end to end run of simulating a single light curve from multiple surveys."""
    # The first survey points at two locations in the sky in the "g" and "r" bands.
    obsdata1 = {
        "time": [0.0, 1.0, 2.0, 3.0],
        "ra": [0.0, 0.0, 180.0, 180.0],
        "dec": [10.0, 10.0, -10.0, -10.0],
        "filter": ["g", "r", "g", "r"],
        "zp": [0.4, 0.5, 0.6, 0.7],
        "seeing": [1.12, 1.12, 1.12, 1.12],
        "skybrightness": [20.0, 20.0, 20.0, 20.0],
        "exptime": [29.2, 29.2, 29.2, 29.2],
        "nexposure": [2, 2, 2, 2],
        "custom_col": [1, 1, 1, 1],
    }
    obstable1 = OpSim(obsdata1)
    passband_group1 = PassbandGroup.from_preset(
        preset="LSST",
        table_dir=test_data_dir / "passbands",
        filters=["g", "r"],
    )

    # The second survey points at two locations on the sky in the "r" and "z" bands.
    obsdata2 = {
        "time": [0.5, 1.5, 2.5, 3.5],
        "ra": [0.0, 90.0, 0.0, 90.0],
        "dec": [10.0, -10.0, 10.0, -10.0],
        "filter": ["r", "z", "r", "z"],
        "zp": [0.05, 0.1, 0.2, 0.3],
        "seeing": [1.12, 1.12, 1.12, 1.12],
        "skybrightness": [20.0, 20.0, 20.0, 20.0],
        "exptime": [29.2, 29.2, 29.2, 29.2],
        "nexposure": [2, 2, 2, 2],
    }
    obstable2 = OpSim(obsdata2)
    passband_group2 = PassbandGroup.from_preset(
        preset="LSST",
        table_dir=test_data_dir / "passbands",
        filters=["r", "z"],
    )

    # Create a constant SED model with known brightnesses and RA, dec values that
    # match the (0.0, 10.0) pointing.
    model = ConstantSEDModel(brightness=100.0, t0=0.0, ra=0.0, dec=10.0, redshift=0.0, node_label="source")
    results = simulate_lightcurves(
        model,
        1,
        [obstable1, obstable2],
        [passband_group1, passband_group2],
        obstable_save_cols=["zp", "custom_col"],
    )
    assert len(results) == 1
    assert results["nobs"][0] == 4

    # Check that the light curve was simulated correctly, including saving the zeropoint information
    # from each ObsTable.
    lightcurve = results["lightcurve"][0]
    assert np.allclose(lightcurve["mjd"], np.array([0.0, 1.0, 0.5, 2.5]))
    assert np.allclose(lightcurve["zp"], np.array([0.4, 0.5, 0.05, 0.2]))
    assert np.array_equal(lightcurve["filter"], np.array(["g", "r", "r", "r"]))
    assert np.array_equal(lightcurve["survey_idx"], np.array([0, 0, 1, 1]))

    # The custom column should only exist for observations from one of the surveys.
    assert np.all(lightcurve["custom_col"][0:2] == 1)
    assert np.all(np.isnan(lightcurve["custom_col"][2:4]))

    # We fail if we pass in lists of different lengths.
    with pytest.raises(ValueError):
        simulate_lightcurves(
            model,
            1,
            [obstable1, obstable2],
            passband_group1,
        )

    # We fail if we try to use a bandflux only model with multiple surveys.
    model2 = StaticBandfluxModel({"g": 1.0, "i": 1.0, "r": 1.0, "y": 1.0, "z": 1.0})
    with pytest.raises(ValueError):
        simulate_lightcurves(
            model2,
            1,
            [obstable1, obstable2],
            [passband_group1, passband_group2],
        )


def test_compute_noise_free_lightcurves_single(test_data_dir):
    """Test computing noise free light curves for a set of times and filters."""
    # Load the passband data for the griz filters only.
    passband_group = PassbandGroup.from_preset(
        preset="LSST",
        table_dir=test_data_dir / "passbands",
        filters=["g", "r", "i", "z"],
    )

    # Create a step model that changes brightness at t=10.0
    model = StepModel(brightness=100.0, t0=10.0, t1=30.0, node_label="source")
    graph_state = model.sample_parameters(num_samples=1)

    lightcurves = compute_single_noise_free_lightcurve(
        model,
        graph_state,
        passband_group,
        rest_frame_phase_min=-40.0,  # 40 days before t0
        rest_frame_phase_max=60.0,  # 60 days after t0
        rest_frame_phase_step=1.0,  # 1 sample per day
    )
    rest_phase = np.arange(-40.0, 60.0, 1.0)
    obs_times = rest_phase + 10.0
    assert np.array_equal(lightcurves["times"], obs_times)
    assert np.array_equal(lightcurves["rest_phase"], rest_phase)

    mask = (obs_times >= 10.0) & (obs_times <= 30.0)
    for filter in ["g", "r", "i", "z"]:
        assert filter in lightcurves
        bandfluxes = lightcurves[filter]
        assert len(bandfluxes) == len(obs_times)
        assert np.allclose(bandfluxes[~mask], 0.0)
        assert np.allclose(bandfluxes[mask], 100.0)


def test_compute_noise_free_lightcurves_multiple(test_data_dir):
    """Test computing noise free light curves for a set of times and filters
    and multiple sample states"""
    # Load the passband data for the griz filters only.
    passband_group = PassbandGroup.from_preset(
        preset="LSST",
        table_dir=test_data_dir / "passbands",
        filters=["g", "r", "i", "z"],
    )

    # Create a step model that changes brightness at t0
    start_times = [10.0, 15.0, 20.0, 25.0]
    end_times = [30.0, 30.0, 40.0, 40.0]
    model = StepModel(
        brightness=100.0,
        t0=GivenValueList(start_times),
        t1=GivenValueList(end_times),
        ra=0.0,
        dec=0.0,
        redshift=0.0,
        node_label="source",
    )
    graph_state = model.sample_parameters(num_samples=4)

    lightcurves = compute_noise_free_lightcurves(
        model,
        graph_state,
        passband_group,
        rest_frame_phase_min=-40.0,  # 40 days before t0
        rest_frame_phase_max=60.0,  # 60 days after t0
        rest_frame_phase_step=1.0,  # 1 sample per day
    )
    assert len(lightcurves) == 4

    rest_phase = np.arange(-40.0, 60.0, 1.0)
    for idx in range(4):
        assert lightcurves["id"][idx] == idx
        lc_df = lightcurves["lightcurve"][idx]
        assert np.array_equal(lc_df["rest_phase"], rest_phase)
        assert np.array_equal(lc_df["times"], rest_phase + start_times[idx])

        mask = (lc_df["times"] >= start_times[idx]) & (lc_df["times"] <= end_times[idx])
        for filter in ["g", "r", "i", "z"]:
            assert filter in lc_df
            bandfluxes = lc_df[filter]
            assert len(bandfluxes) == len(lc_df["times"])
            assert np.allclose(bandfluxes[~mask], 0.0)
            assert np.allclose(bandfluxes[mask], 100.0)


def test_saturation_mags_initialization(test_data_dir):
    """Check initialization of saturation thresholds."""
    # A FakeObsTable should have no saturation thresholds by default.
    fake_obs = FakeObsTable(
        pd.DataFrame(
            {
                "time": [0.0, 1.0],
                "ra": [0.0, 0.0],
                "dec": [0.0, 0.0],
                "filter": ["g", "r"],
            }
        ),
        noise_strategy="exhaustive",
        zp_per_band={"g": 26.0, "r": 27.0},
        exptime=30.0,
        fwhm_px={"g": 2.5, "r": 3.1},
        nexposure=2,
        radius=100.0,
        read_noise=5.0,
        sky_bg_electrons={"g": 150.0, "r": 140.0},
        survey_name="MY_SURVEY",
    )
    assert fake_obs._saturation_mags is None

    # Opsim saturation thresholds should be set by default.
    opsim_db = OpSim.from_db(test_data_dir / "opsim_small.db")
    assert opsim_db._saturation_mags is not None
    assert isinstance(opsim_db._saturation_mags, dict)
    for filter in ["u", "g", "r", "i", "z", "y"]:
        assert filter in opsim_db._saturation_mags

    # For now, the ZTF table uses a single estimate for all filters.
    ztf_table_data = create_random_ztf_obs_data(100)
    ztf_obs_table = ZTFObsTable(table=ztf_table_data)
    assert ztf_obs_table._saturation_mags is not None
    assert isinstance(ztf_obs_table._saturation_mags, dict)
    for filter in ["g", "r", "i"]:
        assert filter in ztf_obs_table._saturation_mags


def test_simulate_with_saturation_mags_as_none(test_data_dir):
    """Test an end to end run of simulating a single light curve with no saturation thresholds."""
    # Set up obs table (with no saturation thresholds set)
    obs_table_values = {
        "time": np.array([0.0, 1.0, 2.0, 3.0, 4.0]),
        "ra": np.array([15.0, 30.0, 15.0, 0.0, 60.0]),
        "dec": np.array([-10.0, -5.0, 0.0, 5.0, 10.0]),
        "filter": np.array(["r", "g", "r", "i", "g"]),
    }
    pdf = pd.DataFrame(obs_table_values)
    zp_per_band = {"g": 26.0, "r": 27.0, "i": 28.0}

    ops_data = FakeObsTable(
        pdf,
        noise_strategy="exhaustive",
        zp_per_band=zp_per_band,
        exptime=60.0,
        fwhm_px={"g": 2.5, "r": 3.1, "i": 1.9},
        nexposure=200,
        radius=100.0,
        read_noise=5.0,
        sky_bg_electrons={"g": 150.0, "r": 140.0, "i": 155.0},
        survey_name="MY_SURVEY",
    )
    assert ops_data._saturation_mags is None

    # Set up model
    source = ConstantSEDModel(
        brightness=50_000.0,
        t0=0.0,
        ra=GivenValueList(obs_table_values["ra"]),
        dec=GivenValueList(obs_table_values["dec"]),
        redshift=0.0,
        node_label="source",
    )

    # Simulate
    passband_group = PassbandGroup.from_preset(
        preset="LSST",
        table_dir=test_data_dir / "passbands",
        filters=["g", "r", "i", "z"],
    )
    results = simulate_lightcurves(
        source,
        1,
        ops_data,
        passband_group,
        obstable_save_cols=["observationId", "zp_nJy"],
        param_cols=["source.brightness"],
    )
    assert len(results) == 1

    # Check that the "is_saturated" column is False for all observations.
    lightcurve = results["lightcurve"][0]
    assert "is_saturated" in lightcurve.columns
    assert not np.any(lightcurve["is_saturated"])


def test_simulate_with_default_saturation_mags_values(test_data_dir):
    """Test an end to end run of simulating a single light curve with default saturation thresholds."""
    # Load the OpSim data.
    opsim_db = OpSim.from_db(test_data_dir / "opsim_small.db")

    # Load the passband data for the g and r filters only.
    passband_group = PassbandGroup.from_preset(
        preset="LSST",
        table_dir=test_data_dir / "passbands",
        filters=["g", "r"],
    )

    # Create a constant SED model with known RA and dec values that match the opsim.
    # Note the brightness is set very high to ensure saturation threshold is surpassed.
    source = ConstantSEDModel(
        brightness=(2.0e12),  # Sufficiently bright to ensure saturation
        t0=0.0,
        ra=GivenValueList(opsim_db["ra"].values[0:5]),
        dec=GivenValueList(opsim_db["dec"].values[0:5]),
        redshift=0.0,
        node_label="source",
    )

    results = simulate_lightcurves(
        source,
        1,
        opsim_db,
        passband_group,
        obstable_save_cols=["observationId", "zp_nJy"],
        param_cols=["source.brightness"],
    )
    assert len(results) == 1

    # Check that every flux value is below the saturation threshold for its filter.
    lightcurve = results["lightcurve"][0]
    opsim_sat_thresholds_njy = {band: mag2flux(mag) for band, mag in opsim_db._saturation_mags.items()}

    for filter in ["g", "r"]:
        mask = lightcurve["filter"] == filter
        fluxes = lightcurve["flux"][mask]
        assert np.all(fluxes <= opsim_sat_thresholds_njy[filter])

    # Check the flux_error values are non-zero.
    flux_errors = lightcurve["fluxerr"]
    assert np.all(flux_errors > 0.0)

    # Check that the "is_saturated" column is True for all observations.
    assert "is_saturated" in lightcurve.columns
    assert np.all(lightcurve["is_saturated"])


def test_simulate_with_custom_saturation_mags(test_data_dir):
    """Test an end to end run of simulating a single light curve with custom saturation thresholds."""
    # Set up obs table
    obs_table_values = {
        "time": np.array([0.0, 1.0, 2.0, 3.0, 4.0]),
        "ra": np.array([15.0, 30.0, 15.0, 0.0, 60.0]),
        "dec": np.array([-10.0, -5.0, 0.0, 5.0, 10.0]),
        "filter": np.array(["r", "g", "r", "i", "g"]),
    }
    pdf = pd.DataFrame(obs_table_values)
    zp_per_band = {"g": 26.0, "r": 27.0, "i": 28.0}

    # Saturation thresholds are given in magnitudes.
    toy_sat_thresholds = {
        "g": 17.5,
        "r": 18.0,
        "i": 17.0,
        "z": 17.2,
    }

    ops_data = FakeObsTable(
        pdf,
        noise_strategy="exhaustive",
        zp_per_band=zp_per_band,
        exptime=60.0,
        fwhm_px={"g": 2.5, "r": 3.1, "i": 1.9},
        nexposure=200,
        radius=100.0,
        read_noise=5.0,
        sky_bg_electrons={"g": 150.0, "r": 140.0, "i": 155.0},
        survey_name="MY_SURVEY",
        saturation_mags=toy_sat_thresholds,
    )

    # Set up model
    source = ConstantSEDModel(
        brightness=5_000_000_000_000.0,  # Sufficiently bright to ensure saturation
        t0=0.0,
        ra=GivenValueList(obs_table_values["ra"]),
        dec=GivenValueList(obs_table_values["dec"]),
        redshift=0.0,
        node_label="source",
    )

    # Simulate
    passband_group = PassbandGroup.from_preset(
        preset="LSST",
        table_dir=test_data_dir / "passbands",
        filters=["g", "r", "i", "z"],
    )
    results = simulate_lightcurves(
        source,
        1,
        ops_data,
        passband_group,
        obstable_save_cols=["observationId", "zp_nJy"],
        param_cols=["source.brightness"],
    )
    assert len(results) == 1

    # Check that every flux value is below the saturation threshold for its filter.
    lightcurve = results["lightcurve"][0]
    toy_sat_thresholds_njy = {band: mag2flux(mag) for band, mag in toy_sat_thresholds.items()}
    for filter in ["g", "r", "i", "z"]:
        mask = lightcurve["filter"] == filter
        fluxes = lightcurve["flux"][mask]
        assert np.all(fluxes <= toy_sat_thresholds_njy[filter])

    # Check the flux_error values are non-zero.
    flux_errors = lightcurve["fluxerr"]
    fluxes = lightcurve["flux"]
    assert np.all(flux_errors > 0.0)

    # Check that the "is_saturated" column is True for all observations.
    assert "is_saturated" in lightcurve.columns
    assert np.all(lightcurve["is_saturated"])



"""The top-level module for survey related data, such as pointing and noise
information. ObsTable class is a base class with specific implementations
for different survey data, such as Rubin and ZTF."""

import logging
import sqlite3
import warnings
from pathlib import Path

import astropy.units as u
import numpy as np
import pandas as pd
from astropy.coordinates import Latitude, Longitude
from mocpy import MOC
from regions import Region
from scipy.spatial import KDTree
from tqdm import tqdm

from lightcurvelynx.astro_utils.coordinate_utils import dedup_coords, ra_dec_to_cartesian
from lightcurvelynx.astro_utils.detector_footprint import DetectorFootprint
from lightcurvelynx.astro_utils.mag_flux import mag2flux

logger = logging.getLogger(__name__)


class ObsTable:
    """A wrapper class around the observations table with helper computation functions and
    cached data for efficiency.

    Parameters
    ----------
    table : dict or pandas.core.frame.DataFrame
        The table with all the survey information. Metadata can be included in the
        "lightcurvelynx_survey_data" entry of the attributes dictionary.
    colmap : dict, optional
        A mapping of standard column names to their names in the input table.
        For example, in Rubin's OpSim we might have the column "observationStartMJD"
        which maps to "time". In that case we would have an entry with key="time"
        and value="observationStartMJD".
    detector_footprint : astropy.regions.SkyRegion, Astropy.regions.PixelRegion, or
        DetectorFootprint, optional
        The footprint object for the instrument's detector. If None, no footprint
        filtering is done. Default is None.
    wcs : astropy.wcs.WCS, optional
        The WCS for the footprint. Either this or pixel_scale must be provided if
        a footprint is provided as a Astropy region.
    saturation_mags : dict, optional
        A dictionary mapping filter names to their saturation thresholds in magnitudes.
        The filters provided must match those in the table. If not provided,
        saturation effects will not be applied.
    **kwargs : dict
        Additional keyword arguments to pass to the constructor. This can include
        overrides of any of the survey values.

    Attributes
    ----------
    survey_values : dict, optional
        A mapping for constant values for the survey used in various computations, such
        as readout noise and dark current.
    filters : np.ndarray
        The unique filters in the survey table (if provided).
    _table : pandas.core.frame.DataFrame
        The table with all the observation information mapped to standard column names.
    _colmap : dict
        A mapping of standard column names to their names in the input table.
    _inv_colmap : dict
        A dictionary mapping the custom column names back to the standard names.
    _kd_tree : scipy.spatial.KDTree or None
        A kd_tree of the survey pointings for fast spatial queries. We use the scipy
        kd-tree instead of astropy's functions so we can directly control caching.
    _detector_footprint : DetectorFootprint, optional
        The footprint object for the instrument's detector. If None, no footprint
        filtering is done. Default is None.
    _wacs : astropy.wcs.WCS, optional
        The WCS for the footprint.
    _saturation_mags : dict, optional
        The saturation thresholds in magnitudes for each filter. If unspecified, an
        instrument-specific default will be used, if available.
    """

    _required_columns = ["ra", "dec", "time"]

    # Default survey values. These are all None for the abstract base class.
    _default_survey_values = {
        "dark_current": None,
        "ext_coeff": None,
        "pixel_scale": None,
        "radius": None,
        "read_noise": None,
        "zp_per_sec": None,
        "survey_name": "Unknown",
    }

    def __init__(
        self,
        table,
        *,
        colmap=None,
        detector_footprint=None,
        wcs=None,
        saturation_mags=None,
        **kwargs,
    ):
        # Create a copy of the table.
        if isinstance(table, dict):
            self._table = pd.DataFrame(table)
        else:
            self._table = table.copy()

        # Remap the columns to standard names. Start with the existing names (from the table)
        # and overwrite anything provided by the column map. Save the inverse mapping.
        name_map = {col: col for col in self._table.columns}
        self._inv_colmap = {}
        self._colmap = colmap if colmap is not None else {}
        if colmap is not None:
            for key, value in colmap.items():
                if value in name_map:
                    # Check for collisions (mapping a column to an existing column)
                    if key in self._table.columns and key != value:
                        raise ValueError(f"Trying to map {value} to {key}, but {key} is already a column.")

                    # Add this entry to the list of column names that need to be remapped.
                    name_map[value] = key

                # Save the inverse mapping as well
                self._inv_colmap[value] = key
        self._table.rename(columns=name_map, inplace=True)

        # Check that we have the required columns.
        for col in self._required_columns:
            if col not in self._table.columns:
                raise KeyError(f"Missing required column: {col}")
        logger.debug(f"ObsTable initialized with columns: {self._table.columns.tolist()}")

        # Save the survey values, with table metadata and keyword arguments overwriting the defaults.
        self.survey_values = self._default_survey_values.copy()
        if "lightcurvelynx_survey_data" in self._table.attrs:
            metadata = self._table.attrs["lightcurvelynx_survey_data"]
            if not isinstance(metadata, dict):
                raise TypeError("Got unexpected type for lightcurvelynx_survey_data")
            for key, value in metadata.items():
                self.survey_values[key] = value
        for key, value in kwargs.items():
            self.survey_values[key] = value
        logger.debug(f"ObsTable survey values: {self.survey_values}")

        self.filters = np.unique(self._table["filter"]) if "filter" in self._table.columns else np.array([])

        # If we are not given zero point data, try to derive it from the other columns.
        if "zp" not in self:
            self._assign_zero_points()

        # Save the saturation thresholds if provided.
        self._saturation_mags = saturation_mags

        # Build the kd-tree.
        self._kd_tree = None
        self._build_kd_tree()

        # Create the footprint if one is provided.
        self._wcs = wcs
        if isinstance(detector_footprint, Region):
            pixel_scale = self.survey_values.get("pixel_scale", None)
            detector_footprint = DetectorFootprint(detector_footprint, wcs=wcs, pixel_scale=pixel_scale)
        self._detector_footprint = detector_footprint

        # Check that the radius is valid for the given footprint (if it exists).
        if self._detector_footprint is not None:
            fp_radius = self._detector_footprint.compute_radius()
            curr_radius = self.survey_values.get("radius", None)
            if curr_radius is None:
                self.survey_values["radius"] = fp_radius
            elif curr_radius < fp_radius:
                logger.info(
                    f"Provided radius {curr_radius} is smaller than footprint radius {fp_radius}. "
                    "Using the footprint radius instead."
                )
                self.survey_values["radius"] = fp_radius
            else:
                logger.debug(
                    f"Provided radius {curr_radius} is larger than footprint radius {fp_radius}. "
                    "Using the provided radius."
                )

    def __len__(self):
        return len(self._table)

    def __getitem__(self, key):
        """Access the underlying observation table by column or parameter name. This will
        return either a full column from the table or a survey parameter value.
        """
        if key in self._table.columns:
            return self._table[key]
        if key in self._inv_colmap and self._inv_colmap[key] in self._table.columns:
            return self._table[self._inv_colmap[key]]
        if key in self.survey_values:
            return self.survey_values[key]
        raise KeyError(f"Column or parameter not found: {key}")

    def __contains__(self, key):
        """Check if a column exists in the survey table or a parameter in the parameter table."""
        if key in self._table.columns:
            return True
        if key in self._inv_colmap and self._inv_colmap[key] in self._table.columns:
            return True
        if key in self.survey_values and self.survey_values[key] is not None:
            return True
        return False

    def clear_detector_footprint(self):
        """Clear the detector footprint, so no footprint filtering is done."""
        self._detector_footprint = None

    def get_value_per_row(self, key, *, indices=None, default=None):
        """Get the values for each row from the table or survey values (defaults).

        Parameters
        ----------
        key : str
            The name of the column to retrieve.
        indices : numpy.ndarray, optional
            The indices of the rows for which to retrieve values. If None, retrieve all rows.
            Default: None
        default : any, optional
            The default value to use if the key is not found in the table or survey values.
            This can be None to indicate missing values. Default: None

        Returns
        -------
        numpy.ndarray
            The values for each row in the table.
        """
        if indices is None:
            indices = np.arange(len(self._table))

        # Prioritize columns that are in the table.
        if key in self._table.columns:
            return self._table[key][indices].to_numpy()
        if key in self._inv_colmap and self._inv_colmap[key] in self._table.columns:
            return self._table[self._inv_colmap[key]][indices].to_numpy()

        # Otherwise fall back to the survey values if they are defined.
        value = self.survey_values.get(key, None)
        if value is None:
            return np.full((len(indices),), default)
        if isinstance(value, float | int):
            # Use the same value for all rows.
            return np.full((len(indices),), value)
        if isinstance(value, dict):
            # Map the values for each filter to the rows in the table.
            result = np.zeros(len(indices), dtype=float)
            for fil, val in value.items():
                if fil not in self.filters:
                    raise ValueError(f"Dictionary for '{key}' does not have a value for filter '{fil}'")
                result[self._table["filter"][indices] == fil] = val
            return result
        raise TypeError(f"Unsupported type for '{key}': {type(value)}")

    def safe_get_survey_value(self, key):
        """Get a survey value by key, checking that it is not None.

        Parameters
        ----------
        key : str
            The key of the survey value to retrieve.
        """
        value = self.survey_values.get(key, None)
        if value is None:
            raise ValueError(
                f"Survey value for {key} is not defined. This should be set when creating the object."
            )
        return value

    @property
    def radius(self):
        """Return the radius if it exists."""
        return self.survey_values.get("radius", None)

    @radius.setter
    def radius(self, new_val):
        """Create a setter for radius."""
        if new_val <= 0:
            raise ValueError(f"Invalid radius: {new_val}")
        if self._detector_footprint is not None:
            fp_radius = self._detector_footprint.compute_radius()
            if new_val < fp_radius:
                warnings.warn(
                    f"Provided radius {new_val} is smaller than footprint radius {fp_radius}. "
                    "This might lead to unexpected results."
                )
        self.survey_values["radius"] = new_val

    @property
    def columns(self):
        """Get the column names."""
        return self._table.columns

    @classmethod
    def from_db(cls, filename, sql_query="SELECT * FROM observations", **kwargs):
        """Create an ObsTable object from the data in an db file. Reads data matching
        what is produced by write_db (and matching the RubinOpsim table).

        Parameters
        ----------
        filename : str
            The name of the db file.
        sql_query : str
            The SQL query to use when loading the table.
            Default: "SELECT * FROM observations"
        kwargs : dict, optional
            Additional keyword arguments to pass to the Survey constructor.

        Returns
        -------
        ObsTable
            A table with all of the pointing data.

        Raise
        -----
        FileNotFoundError if the file does not exist.
        ValueError if unable to load the table.
        """
        if not Path(filename).is_file():
            raise FileNotFoundError(f"db file {filename} not found.")
        con = sqlite3.connect(f"file:{filename}?mode=ro", uri=True)

        # Read the table.
        try:
            survey_data = pd.read_sql_query(sql_query, con)
        except Exception:
            raise ValueError("Database read failed.") from None

        # Close the connection.
        con.close()

        return cls(survey_data, **kwargs)

    @classmethod
    def from_parquet(cls, filename):
        """Create an ObsTable object from a parquet file.

        Parameters
        ----------
        filename : str
            The name of the parquet file to read.

        Returns
        -------
        ObsTable
            A table with all of the pointing data.
        """
        if not Path(filename).is_file():
            raise FileNotFoundError(f"File {filename} not found.")
        survey_data = pd.read_parquet(filename)
        return cls(survey_data)

    def estimate_coverage(self, *, radius=None, max_depth=12, use_footprint=False):
        """Estimate the sky coverage of the observations in the ObsTable. This is an
        approximate calculation based on a constructed MOC at a given depth.

        Parameters
        ----------
        radius : float, optional
            The radius to use for each image (in degrees). Only used if use_footprint
            is False. If None, the radius from the survey values will be used.
        max_depth : int, optional
            The maximum depth of the MOC. Default is 12.
        use_footprint : bool, optional
            Whether to use the detector footprint to build the MOC. If True, the
            footprint will be used to compute the MOC regions for each pointing.
            If False, a simple cone with the given radius will be used.

        Returns
        -------
        coverage : float
            The estimated sky coverage in square degrees.
        """
        moc = self.build_moc(radius=radius, max_depth=max_depth, use_footprint=use_footprint)
        coverage = moc.sky_fraction * 41253.0  # Approximate sky area in deg^2
        return coverage

    def build_moc(
        self,
        *,
        duplicate_threshold=100.0 / 3600.0,
        max_depth=10,
        radius=None,
        use_footprint=False,
    ):
        """Build a Multi-Order Coverage Map from the regions in the data set.

        The MOCs can be either from simple cones around each pointing or from the
        detector footprint if available. The code does not currently support rotation
        when dealing with detector footprints.

        Note
        ----
        This can be **very** slow for large ObsTables or high max_depth values, especially
        when using detector footprints.

        Parameters
        ----------
        duplicate_threshold : float, optional
            The threshold to use for identifying duplicate pointings (in degrees).
            Default is 100.0 / 3600.0 degrees = 100 arcseconds.
        max_depth : int, optional
            The maximum depth of the MOC. Default is 10.
        radius : float, optional
            The radius to use for each image (in degrees). Only used if use_footprint
            is False. If None, the radius from the survey values will be used.
        use_footprint : bool, optional
            Whether to use the detector footprint to build the MOC. If True, the
            footprint will be used to compute the MOC regions for each pointing.
            If False, a simple cone with the given radius will be used.

        Returns
        -------
        MOC
            The Multi-Order Coverage Map constructed from the data set.
        """
        logger.debug(
            f"Building MOC from ObsTable data: Depth={max_depth}, use_footprint={use_footprint}, "
            f"duplicate_threshold={duplicate_threshold}"
        )

        # Deduplicate near-matching pointings to save computation time.
        ra = self._table["ra"].to_numpy()
        dec = self._table["dec"].to_numpy()
        if duplicate_threshold > 0.0:
            ra, dec, _ = dedup_coords(ra, dec, threshold=duplicate_threshold)
            logger.debug(f"Filtered {len(self._table) - len(ra)} duplicate pointings for MOC construction.")
        logger.debug(f"Building MOC from {len(ra)} unique pointings.")

        if not use_footprint or self._detector_footprint is None:
            radius = radius if radius is not None else self.survey_values.get("radius", None)
            if radius is None:
                raise ValueError("Radius must be provided for MOC construction or as a default. Got None.")

            moc = MOC.from_cones(
                lon=Longitude(ra, unit="deg"),
                lat=Latitude(dec, unit="deg"),
                radius=radius * u.deg,
                max_depth=max_depth,
                delta_depth=0,
                union_strategy="large_cones",
            )
        else:
            # The combination of arbitrary rotations and mocpy does not currently work together
            # (e.g. a rotated rectangle of 90.1 degrees will fail, because the max rotation is capped
            # at pi radians). So we ignore rotation for now.
            if "rotation" in self._table.columns:
                warnings.warn(
                    "MOC construction with footprint does not support rotation. Ignoring rotation values."
                )

            moc = None
            for curr_ra, curr_dec in tqdm(
                zip(ra, dec, strict=False), total=len(ra), desc="Evaluating Region"
            ):
                sky_region, _ = self._detector_footprint.compute_sky_region(curr_ra, curr_dec)
                new_moc = MOC.from_astropy_regions(sky_region, max_depth=max_depth)
                moc = new_moc if moc is None else MOC.union(moc, new_moc)

        return moc

    def _build_kd_tree(self):
        """Construct the KD-tree from the ObsTable."""
        # Convert the pointings to Cartesian coordinates on a unit sphere.
        x, y, z = ra_dec_to_cartesian(self._table["ra"].to_numpy(), self._table["dec"].to_numpy())
        cart_coords = np.array([x, y, z]).T

        # Construct the kd-tree.
        self._kd_tree = KDTree(cart_coords)

    def _assign_zero_points(self):
        """Assign instrumental zero points in nJy to the data table.

        Default implementation does not produce a zeropoint column. Subclasses
        should override this method with a survey specific computation.
        """
        pass

    def add_column(self, colname, values, *, overwrite=False):
        """Add a column to the current data table.

        Parameters
        ----------
        colname : str
            The name of the new column.
        values : int, float, str, list, or numpy.ndarray
            The value(s) to add.
        overwrite : bool
            Overwrite the column is it already exists.
            Default: False
        """
        if colname in self._table.columns and not overwrite:
            raise KeyError(f"Column {colname} already exists.")

        # If the input is a scalar, turn it into an array of the correct length
        if np.isscalar(values):
            values = np.full((len(self._table)), values)
        self._table[colname] = values

    def write_db(self, filename, *, tablename="observations", overwrite=False):
        """Write out the observation table as a database to a given SQL table.

        Parameters
        ----------
        filename : str
            The name of the db file.
        tablename : str
            The table to which to write.
            Default: "observations"
        overwrite : bool
            Overwrite the existing DB file.
            Default: False

        Raise
        -----
        FileExistsError if the file already exists and overwrite is False.
        """
        if_exists = "replace" if overwrite else "fail"

        con = sqlite3.connect(filename)
        try:
            self._table.to_sql(tablename, con, if_exists=if_exists)
        except Exception:
            raise ValueError("Database write failed.") from None

        con.close()

    def write_parquet(self, filename, *, overwrite=False):
        """Write out the observation table as a parquet file.

        Parameters
        ----------
        filename : str
            The name of the parquet file.
        overwrite : bool
            Overwrite the existing parquet file.
            Default: False

        Raise
        -----
        FileExistsError if the file already exists and overwrite is False.
        """
        if not overwrite and Path(filename).is_file():
            raise FileExistsError(f"File {filename} already exists.")

        # Save all the survey data as metadata.
        self._table.attrs["lightcurvelynx_survey_data"] = self.survey_values
        self._table.to_parquet(filename)

    def time_bounds(self):
        """Returns the min and max times for all observations in the ObsTable.

        Returns
        -------
        t_min, t_max : float, float
            The min and max times for all observations in the ObsTable.
        """
        t_min = self._table["time"].min()
        t_max = self._table["time"].max()
        return t_min, t_max

    def filter_rows(self, rows):
        """Filter the rows in the ObsTable to only include those indices that are provided
        in a list of row indices (integers) or marked True in a mask.

        Parameters
        ----------
        rows : numpy.ndarray
            Either a Boolean array of the same length as the table or list of integer
            row indices to keep.

        Returns
        -------
        self : ObsTable
            The filtered ObsTable object.
        """
        # Check if we are dealing with a mask of a list of indices.
        rows = np.asarray(rows)
        if rows.dtype == bool:
            if len(rows) != len(self._table):
                raise ValueError(
                    f"Mask length mismatch. Expected {len(self._table)} rows, but found {len(rows)}."
                )
            mask = rows
        else:
            mask = np.full((len(self._table),), False)
            mask[rows] = True

        # Filter the rows in-place and build a new kd-tree.
        self._table = self._table[mask]
        self._kd_tree = None
        self._build_kd_tree()

        return self

    def is_observed(self, query_ra, query_dec, *, radius=None, t_min=None, t_max=None):
        """Check if the query point(s) fall within the field of view of any
        pointing in the ObsTable.

        Parameters
        ----------
        query_ra : float or numpy.ndarray
            The query right ascension (in degrees).
        query_dec : float or numpy.ndarray
            The query declination (in degrees).
        radius : float or None, optional
            The angular radius of the observation (in degrees).
        t_min : float or None, optional
            The minimum time (in MJD) for the observations to consider.
            If None, no time filtering is applied.
        t_max : float or None, optional
            The maximum time (in MJD) for the observations to consider.
            If None, no time filtering is applied.

        Returns
        -------
        seen : bool or list[bool]
            Depending on the input, this is either a single bool to indicate
            whether the query point is observed or a list of bools for an array
            of query points.
        """
        inds = self.range_search(query_ra, query_dec, radius=radius, t_min=t_min, t_max=t_max)
        if np.isscalar(query_ra):
            return len(inds) > 0
        return [len(entry) > 0 for entry in inds]

    def range_search(self, query_ra, query_dec, *, radius=None, t_min=None, t_max=None):
        """Return the indices of the pointings that fall within the field
        of view of the query point(s).

        Parameters
        ----------
        query_ra : float or numpy.ndarray
            The query right ascension (in degrees).
        query_dec : float or numpy.ndarray
            The query declination (in degrees).
        radius : float or None, optional
            The angular radius of the observation (in degrees). If None
            uses the default radius for the ObsTable.
        t_min : float, numpy.ndarray or None, optional
            The minimum time (in MJD) for the observations to consider.
            If None, no time filtering is applied.
        t_max : float, numpy.ndarray or None, optional
            The maximum time (in MJD) for the observations to consider.
            If None, no time filtering is applied.

        Returns
        -------
        inds : list[int] or list[numpy.ndarray]
            Depending on the input, this is either a list of indices for a single query point
            or a list of arrays (of indices) for an array of query points.
        """
        if query_ra is None or query_dec is None:
            raise ValueError("Query RA and dec must be provided for range search, but got None.")

        # Fallback to the preset radius if None is provided. Output a warning if a *smaller*
        # radius is provided, since that may lead to unexpected results if a footprint is used.
        if radius is not None:
            if self.radius is not None and radius < self.radius:
                warnings.warn(
                    f"Provided radius {radius} is smaller than the ObsTable radius {self.radius}. "
                    "This may lead to unexpected results if a detector footprint is used."
                )
        else:
            radius = self.survey_values.get("radius", None)
            if radius is None:
                raise ValueError("Radius must be provided for range search or as a default. Got None.")

        # If the points are scalars, make them into length 1 arrays.
        is_scalar = np.isscalar(query_ra) and np.isscalar(query_dec)
        query_ra = np.atleast_1d(query_ra)
        query_dec = np.atleast_1d(query_dec)

        # Confirm the query RA and Dec have the same length.
        if len(query_ra) != len(query_dec):
            raise ValueError("Query RA and Dec must have the same length.")
        if np.any(query_ra == None) or np.any(query_dec == None):  # noqa: E711
            raise ValueError("Query RA and dec cannot contain None.")

        # Transform the query point(s) to 3-d Cartesian coordinate(s).
        x, y, z = ra_dec_to_cartesian(query_ra, query_dec)
        cart_query = np.array([x, y, z]).T

        # Adjust the angular radius to a cartesian search radius and perform the search.
        adjusted_radius = 2.0 * np.sin(0.5 * np.radians(radius))
        inds = self._kd_tree.query_ball_point(cart_query, adjusted_radius)

        if t_min is not None or t_max is not None:
            num_queries = len(query_ra)
            times = self._table["time"].to_numpy()

            if t_min is None:
                t_min = np.full(num_queries, -np.inf)
            else:
                t_min = np.atleast_1d(t_min)
            if len(t_min) != num_queries:
                raise ValueError(f"t_min must be a scalar or an array of length {num_queries}.")

            if t_max is None:
                t_max = np.full(num_queries, np.inf)
            else:
                t_max = np.atleast_1d(t_max)
            if len(t_max) != num_queries:
                raise ValueError(f"t_max must be a scalar or an array of length {num_queries}.")

            # Run through each list of indices and filter by time. We need to do this
            # iteratively, because the lists can have different lengths.
            for idx, subinds in enumerate(inds):
                if len(subinds) == 0:
                    continue
                time_mask = (times[subinds] >= t_min[idx]) & (times[subinds] <= t_max[idx])
                inds[idx] = np.asarray(subinds)[time_mask]

        # Do a filtering step based on the detector's footprint. We do this after the range search,
        # because it is more expensive (but also more accurate).
        if self._detector_footprint is not None:
            # Extract the RA and dec of the pointings for later use.
            all_ra = self._table["ra"].to_numpy()
            all_dec = self._table["dec"].to_numpy()
            all_rot = None if "rotation" not in self._table.columns else self._table["rotation"].to_numpy()

            for idx, subinds in enumerate(inds):
                num_matches = len(subinds)
                if num_matches == 0:
                    continue  # Nothing to filter.

                match_rot = None if all_rot is None else all_rot[subinds]
                mask = self._detector_footprint.contains(
                    np.full(num_matches, query_ra[idx]),  # The RA coordinate of this query
                    np.full(num_matches, query_dec[idx]),  # The dec coordinate of this query
                    all_ra[subinds],  # The RA coordinates of the pointings (detector positions)
                    all_dec[subinds],  # The dec coordinates of the pointings (detector positions)
                    rotation=match_rot,  # The detector rotation angles (if available)
                )
                inds[idx] = np.asarray(subinds)[mask]

        # If the query was a scalar, we return a single list of indices.
        if is_scalar:
            inds = inds[0]
        return inds

    def get_observations(self, query_ra, query_dec, *, radius=None, t_min=None, t_max=None, cols=None):
        """Return the observation information when the query point falls within
        the field of view of a pointing in the ObsTable.

        Parameters
        ----------
        query_ra : float
            The query right ascension (in degrees).
        query_dec : float
            The query declination (in degrees).
        radius : float or None, optional
            The angular radius of the observation (in degrees). If None
            uses the default radius for the ObsTable.
        t_min : float or None, optional
            The minimum time (in MJD) for the observations to consider.
            If None, no time filtering is applied.
        t_max : float or None, optional
            The maximum time (in MJD) for the observations to consider.
            If None, no time filtering is applied.
        cols : list or str
            A list of the names of columns to extract or a single column name.
            If None returns all the columns.

        Returns
        -------
        results : dict
            A dictionary mapping the given column name to a numpy array of values.
        """
        neighbors = self.range_search(query_ra, query_dec, radius=radius, t_min=t_min, t_max=t_max)

        results = {}
        if cols is None:
            cols = self._table.columns.to_list()
        elif isinstance(cols, str):
            cols = [cols]

        for col in cols:
            # Allow the user to specify either the original or mapped column names,
            # by using the class accessor (__getitem__), instead of the table one.
            if col not in self:
                raise KeyError(f"Unrecognized column name {col}")
            results[col] = self[col][neighbors].to_numpy()
        return results

    def bandflux_error_point_source(self, bandflux, index):
        """Compute observational bandflux error for a point source.

        Parameters
        ----------
        bandflux : array_like of float
            Band bandflux of the point source in nJy.
        index : array_like of int
            The index of the observation in the ObsTable table.

        Returns
        -------
        flux_err : array_like of float
            Simulated bandflux noise in nJy.
        """
        raise NotImplementedError

    def compute_saturation(self, flux, flux_error, index):
        """Apply the saturation limits to a given flux and flux error.

        When a flux value exceeds the saturation limit, it is clipped to the limit and flagged as
        saturated. In these cases, the associated flux_error is increased to account for the offset
        introduced by clipping. The new error is computed as the quadrature sum of the original
        flux_error and the difference between the orginal flux and saturated flux:

            saturated_flux_error = sqrt(flux_error**2 + (flux - saturated_flux)**2)

        For unsaturated points, both flux and flux_error are returned unchanged.

        Parameters
        ----------
        flux : numpy.ndarray of float
            The bandflux in nJy. A size S x T array where S is the
            number of samples in the graph state and T is the number of time points.
        flux_error : numpy.ndarray of float
            The bandflux error in nJy. A size S x T array where S is the
            number of samples in the graph state and T is the number of time points.
        index : array_like of int
            The index of the observation in the ObsTable table.

        Returns
        -------
        tuple of numpy.ndarray
            A tuple with three entries:
            - The saturated flux in nJy. A size S x T array where S is the
                number of samples in the graph state and T is the number of time points.
            - The saturated flux error in nJy. A size S x T array where S is the
                number of samples in the graph state and T is the number of time points.
            - A boolean array indicating which points are saturated. A size S x T array
                where S is the number of samples in the graph state and T is the number of time points.
        """
        if self._saturation_mags is None:
            logger.info("Saturation thresholds not provided. Skipping saturation computation.")
            return flux, flux_error, np.full(flux.shape, False)

        true_flux = np.asarray(flux)
        true_flux_error = np.asarray(flux_error)
        filters = np.asarray(self._table["filter"].iloc[index])

        if len(flux) != len(flux_error) or len(flux) != len(filters):
            raise ValueError("Input arrays must have the same length.")

        # Convert saturation thresholds to nJy.
        saturation_mags_njy = {}
        for filt, mag in self._saturation_mags.items():
            if not isinstance(mag, int | float):
                raise ValueError("Saturation thresholds must be numeric.")
            saturation_mags_njy[filt] = mag2flux(mag)

        # Map the filter list to saturation limits.
        limits = np.array([saturation_mags_njy.get(filt, np.inf) for filt in filters])

        # Calculate the saturated flux and flux error.
        saturated_flux = np.minimum(true_flux, limits)
        saturated_flux_error = np.hypot(true_flux_error, (true_flux - saturated_flux))
        saturated_flux_error = np.where(true_flux <= limits, true_flux_error, saturated_flux_error)

        # Create a flag array to indicate which points are saturated.
        saturation_flags = true_flux > limits

        return saturated_flux, saturated_flux_error, saturation_flags



"""A model for an AGN.

Adapted from https://github.com/RickKessler/SNANA/blob/master/src/gensed_AGN.py
with the authors' permission.
"""

from os import urandom

import numpy as np
from astropy import constants
from citation_compass import cite_function
from scipy import integrate

from lightcurvelynx.base_models import FunctionNode
from lightcurvelynx.consts import ANGSTROM_TO_CM, CGS_FNU_UNIT_TO_NJY, M_SUN_G, PARSEC_TO_CM
from lightcurvelynx.math_nodes.np_random import NumpyRandomFunc
from lightcurvelynx.models.physical_model import SEDModel


def sample_damped_random_walk(times, tau_v, sf_inf, t0, rng=None):
    """Sample a damped random walk.

    Parameters
    ----------
    times : np.ndarray
        A length T array with the times at which to sample the damped random
        walk (in MJD).
    tau_v : np.ndarray
        A length W array with timescale of the damped random walk (in s).
    sf_inf : np.ndarray
        A length W array with structure function at infinity time in magnitude.
    t0 : float
        The initial time moment (in MJD).
    rng : np.random.Generator, optional
        The random number generator to use.
        Default: None

    Returns
    -------
    samples : float
        The sampled value of the damped random walk.
    """
    if rng is None:
        rng = np.random.default_rng()

    if len(sf_inf) != len(tau_v):
        raise ValueError(
            "The arrays sf_inf and tau_v must have the same length. "
            f"Received lengths: sf_inf={len(sf_inf)}, tau_v={len(tau_v)}"
        )
    samples = np.zeros((len(times), len(tau_v)))

    # Set the initial values.
    curr_t = t0
    delta_m = rng.random() * sf_inf

    # Iterate over each time step.
    for idx, time in enumerate(times):
        if time <= curr_t and idx != 0:
            raise ValueError("Times must be monotonically increasing.")

        dt = time - curr_t
        delta_m = delta_m * np.exp(-dt / tau_v) + sf_inf * np.sqrt(1 - np.exp(-2 * dt / tau_v)) * rng.random()
        samples[idx, :] = delta_m

    return samples


class AGN(SEDModel):
    """A model for an AGN.

    Parameterized values include:
      * accretion_rate - The accretion rate (ME_dot) at Eddington luminosity in g/s.
      * blackhole_accretion_rate - The accretion rate of the black hole in g/s.
      * blackhole_mass - The black hole mass in solar mass.
      * edd_ratio - The Eddington ratio.
      * dec - The object's declination in degrees. [from BasePhysicalModel]
      * distance - The object's luminosity distance in pc. [from BasePhysicalModel]
      * inclination - The inclination of the accretion disk in radians (sampled uniformly
        between 0 and pi/2).
      * L_bol - The bolometric luminosity in erg/s.
      * mag_i - The i band absolute magnitude.
      * ra - The object's right ascension in degrees. [from BasePhysicalModel]
      * redshift - The object's redshift. [from BasePhysicalModel]
      * t0 - The t0 of the zero phase, date. [from BasePhysicalModel]

    Parameters
    ----------
    t0 : float
        initial time moment in days.
    blackhole_mass : float
        The black hole mass in solar masses.
    edd_ratio: float
        Eddington ratio
    node_label : str, optional
        The label for the node in the model graph.
        Default: None
    seed : int, optional
        The seed to use for the random number generator.
        Default: None
    **kwargs : dict
        Additional keyword arguments.
    """

    def __init__(self, t0, blackhole_mass, edd_ratio, node_label=None, seed=None, **kwargs):
        # We manually specify "node_label" as a parameter so it does not get
        # passed to the functions nodes below as part of kwargs.
        super().__init__(t0=t0, node_label=node_label, **kwargs)

        if "redshift" not in kwargs:
            raise ValueError("'redshift' parameter is required for the AGN model.")

        if "cosmology" not in kwargs and "distance" not in kwargs:
            raise ValueError(
                "At least one of 'cosmology' (str or astropy.cosmology object) or 'distance' (in pc) "
                "parameters must be provided."
            )

        # Add the parameters for the AGN. t0 already set in BasePhysicalModel.
        self.add_parameter(
            "blackhole_mass",
            blackhole_mass,
            description="The black hole mass in solar mass.",
            **kwargs,
        )
        self.add_parameter(
            "edd_ratio",
            edd_ratio,
            description="The Eddington ratio.",
            **kwargs,
        )
        self.add_parameter(
            "inclination_rad",
            NumpyRandomFunc("uniform", low=0, high=np.pi / 2.0),
            description=(
                "The inclination of the accretion disk in radians (sampled uniformly between 0 and pi/2)."
            ),
            **kwargs,
        )

        # Add the derived parameters using FunctionNodes built from the object's static methods.
        # Each of these will be computed for each sample value of the input parameters.
        self.add_parameter(
            "blackhole_mass_gram",
            FunctionNode(lambda m_sun: m_sun * M_SUN_G, m_sun=self.blackhole_mass),
            description="The black hole mass in grams. Automatically computed from the blackhole_mass.",
            **kwargs,
        )
        self.add_parameter(
            "critical_accretion_rate",
            FunctionNode(self.compute_critical_accretion_rate, blackhole_mass=self.blackhole_mass_gram),
            description=(
                "The critical accretion rate (ME_dot) at Eddington luminosity in g/s. "
                "Automatically computed from the blackhole_mass_gram."
            ),
            **kwargs,
        )
        self.add_parameter(
            "blackhole_accretion_rate",
            FunctionNode(
                self.compute_blackhole_accretion_rate,
                accretion_rate=self.critical_accretion_rate,  # Pull from computed accretion rate
                edd_ratio=self.edd_ratio,  # Pull from sampled ratio
            ),
            description=(
                "The accretion rate of the black hole in g/s. Automatically computed from critical "
                "accretion rate and Eddington ratio."
            ),
            **kwargs,
        )
        self.add_parameter(
            "bolometric_luminosity",
            FunctionNode(
                self.compute_bolometric_luminosity,
                edd_ratio=self.edd_ratio,  # Pull from sampled ratio
                blackhole_mass=self.blackhole_mass_gram,  # Pull from the sampled mass
            ),
            description=(
                "The bolometric luminosity in erg/s. Automatically computed from Eddington "
                "ratio and blackhole mass.",
            ),
            **kwargs,
        )
        self.add_parameter(
            "mag_i",
            FunctionNode(
                self.compute_mag_i,
                bolometric_luminosity=self.bolometric_luminosity,  # Pull from computed value.
            ),
            description=(
                "The i band absolute magnitude. Automatically computed from the bolometric luminosity.",
            ),
            **kwargs,
        )
        self.add_parameter(
            "sf_inf_4000aa",
            FunctionNode(
                self.compute_sf_inf_4000aa,
                mag_i=self.mag_i,
                blackhole_mass=self.blackhole_mass_gram,
            ),
        )
        self.add_parameter(
            "tau_4000aa",
            FunctionNode(
                self.compute_tau_4000aa,
                mag_i=self.mag_i,
                blackhole_mass=self.blackhole_mass_gram,
            ),
        )

        # Set the random number generator for this object.
        if seed is None:
            seed = int.from_bytes(urandom(4), "big")
        self._rng = np.random.default_rng(seed=seed)

    # ------------------------------------------------------------------------
    # --- Static helper methods for computing the derived parameters. --------
    # ------------------------------------------------------------------------

    @staticmethod
    def compute_critical_accretion_rate(blackhole_mass):
        """Compute the critical accretion rate at Eddington luminosity.

        Parameters
        ----------
        blackhole_mass : float
            The black hole mass in g.

        Returns
        -------
        accretion_rate : float
            The accretion rate (ME_dot) at Eddington luminosity in g/s.
        """
        return 1.4e18 * blackhole_mass / M_SUN_G

    @staticmethod
    def compute_blackhole_accretion_rate(accretion_rate, edd_ratio):
        """Compute the accretion rate of the blackhole.

        Parameters
        ----------
        accretion_rate : float
            The accretion rate (ME_dot) at Eddington luminosity in g/s.
        edd_ratio : float
            The Eddington ratio.

        Returns
        -------
        bh_accretion_rate : float
            The accretion rate of the black hole in g/s.
        """
        return accretion_rate * edd_ratio

    @staticmethod
    def compute_bolometric_luminosity(edd_ratio, blackhole_mass):
        """Compute the bolometric luminosity of an AGN.

        Parameters
        ----------
        edd_ratio : float
            The Eddington ratio.
        blackhole_mass : float
            The black hole mass in g.

        Returns
        -------
        bolometric_luminosity : float
            The bolometric luminosity in erg/s.
        """
        return edd_ratio * 1.26e38 * blackhole_mass / M_SUN_G

    @staticmethod
    @cite_function
    def compute_sed_standard_disk(*, Mdot, nu, rin_to_rg, i, d, M):
        """Compute the flux based on a standard disk model.

        References
        ----------
        Lipunova, G., Malanchev, K., Shakura, N. (2018)
        https://doi.org/10.1007/978-3-319-93009-1_1

        All inputs are in CGS.

        Parameters
        ----------
        Mdot : float or np.ndarray
            The accretion rate at the previous time step.
        nu : float or np.ndarray
            The frequency.
        rin_to_rg : float or np.ndarray
            The inner radius of the accretion disk to gravitation radius ratio.
        i : float or np.ndarray
            The inclination in radians.
        d : float or np.ndarray
            The distance.
        M : float or np.ndarray
            The mass of the gravitating center.

        Returns
        -------
        flux : float
            The flux at the given time step.
        """
        # Compute the initial radius of the ring (r_0) and the effective temperature at r_0 (T_0).
        # Lipunova, G., Malanchev, K., Shakura, N. (2018). Page 33 for the main equation
        # DOI https://doi.org/10.1007/978-3-319-93009-1_1
        rin = constants.G.cgs.value * M / constants.c.cgs.value**2 * rin_to_rg
        r_0 = AGN.compute_r_0(rin)
        T_0 = AGN.compute_temp_at_r_0(M, Mdot, rin)

        h = constants.h.cgs.value
        k_B = constants.k_B.cgs.value
        c = constants.c.cgs.value
        # Set minimum x_in value for numerical stability
        x_in = np.maximum(h * nu / (k_B * T_0) * (rin / r_0) ** (3 / 4), 1e-6)
        # Set outer radius to infinity for now
        x_out = np.inf

        # large x in exponetial causes overflow, but 1/inf is zero.
        with np.errstate(over="ignore"):

            def _fun_integr(x):
                return (x ** (5 / 3)) / np.expm1(x)

            integ, _ = np.vectorize(lambda x_in, x_out: integrate.quad(_fun_integr, x_in, x_out))(x_in, x_out)

        return (
            (16 * np.pi)
            / (3 * d**2)
            * np.cos(i)
            * (k_B * T_0 / h) ** (8 / 3)
            * h
            * (nu ** (1 / 3))
            / (c**2)
            * (r_0**2)
            * integ
        ) * CGS_FNU_UNIT_TO_NJY

    @staticmethod
    @cite_function
    def compute_mag_i(bolometric_luminosity):
        """Compute the i band magnitude from the bolometric luminosity.

        References
        ----------
        Shen et al., 2013 - https://adsabs.harvard.edu/full/2013BASI...41...61S

        Parameters
        ----------
        bolometric_luminosity : float
            The bolometric luminosity in erg/s.

        Returns
        -------
        mag_i : float
            The i band magnitude.
        """
        # Adpated from Shen et al., 2013: https://adsabs.harvard.edu/full/2013BASI...41...61S
        return 90 - 2.5 * np.log10(bolometric_luminosity)

    @staticmethod
    @cite_function
    def compute_r_0(r_in):
        """Compute the initial radius of the ring (r_0) in a standard disk model
        given the inner radius.

        References
        ----------
        Lipunova, G., Malanchev, K., Shakura, N. (2018)
        https://doi.org/10.1007/978-3-319-93009-1_1

        Parameters
        ----------
        r_in : float
            The inner radius of the accretion disk.

        Returns
        -------
        r_0 : float
            The initial radius of the ring.
        """
        # Adapted from Lipunova, G., Malanchev, K., Shakura, N. (2018). Page 31
        # DOI https://doi.org/10.1007/978-3-319-93009-1_1
        return (7 / 6) ** 2 * r_in

    @staticmethod
    @cite_function
    def compute_structure_function_at_inf(sf_inf_4000aa, wavelength):
        """Adjust the structure function at infinity time for given wavelengths

        References
        ----------
        Suberlak et al. 2021 - DOI 10.3847/1538-4357/abc698

        Parameters
        ----------
        sf_inf_4000aa : float
            The structure function at infinity time for 4000 angstroms
        wavelength : np.ndarray
            A length W array with wavelength in cm.

        Returns
        -------
        result : np.ndarray
            A length W array with structure function at infinity time in magnitude.
        """
        return sf_inf_4000aa * 10 ** (-0.479) * wavelength / 4000e-8

    @staticmethod
    @cite_function
    def compute_sf_inf_4000aa(mag_i=-23, blackhole_mass=1e9 * M_SUN_G):
        """Compute the structure function at infinity time in magnitude, for 4000 A

        References
        ----------
        Suberlak et al. 2021 - DOI 10.3847/1538-4357/abc698

        Parameters
        ----------
        mag_i : float, optional
            The i band magnitude.
            Default: -23
        blackhole_mass : float, optional
            The black hole mass in g.
            Default: 1e9 * M_SUN_G

        Returns
        -------
        result : np.ndarray
            A length W array with structure function at infinity time
            in magnitude for 4000 angstroms.
        """
        # Equation and parameters for A=-0.51, B=-0.479, C=0.13, and D=0.18
        #  adopted from Suberlak et al. 2021: DOI 10.3847/1538-4357/abc698
        return 10 ** (-0.476 + 0.118 * (mag_i + 23) + 0.118 * np.log10(blackhole_mass / (1e9 * M_SUN_G)))

    @staticmethod
    @cite_function
    def compute_tau_4000aa(mag_i=-23, blackhole_mass=1e9 * M_SUN_G):
        """Compute the timescale (tau_v) at 4000 angstroms.

        References
        ----------
        Suberlak et al. 2021 - DOI 10.3847/1538-4357/abc698

        Parameters
        ----------
        mag_i : float, optional
            The i band magnitude.
            Default: -23
        blackhole_mass : float, optional
            The black hole mass in g.
            Default: 1e9 * M_SUN_G

        Returns
        -------
        tau_v : np.ndarray
            A length W array with the timescale for 4000 angstroms.
        """
        # Equation and parameters for A=2.4, B=0.17, C=0.03, and D=0.21 adopted
        # from Suberlak et al. 2021: DOI 10.3847/1538-4357/abc698
        return 10 ** (2.597 + 0.035 * (mag_i + 23) + 0.141 * np.log10(blackhole_mass / (1e9 * M_SUN_G)))

    @staticmethod
    @cite_function
    def compute_tau(tau_4000aa, wavelength):
        """Adjust the timescale (tau_v) for given wavelengths

        References
        ----------
        Suberlak et al. 2021 - DOI 10.3847/1538-4357/abc698

        Parameters
        ----------
        tau_4000aa : float
            The timescale for 4000 angstroms.
        wavelength : np.ndarray
            A length W array with wavelength in cm.
        """
        return tau_4000aa * 10**0.17 * wavelength / 4000e-8

    @staticmethod
    @cite_function
    def compute_temp_at_r_0(M, Mdot, r_in):
        """Compute the effective temperature at r0. This is the same as the maximum effective
        temperature at the disc surface (Tmax).

        References
        ----------
        Lipunova, G., Malanchev, K., Shakura, N. (2018)
        https://doi.org/10.1007/978-3-319-93009-1_1

        Parameters
        ----------
        M : float
            The mass of the gravitating centre.
        Mdot : float
            The accretion rate at the previous time step.
        r_in : float
            The inner radius of the accretion disk.

        Returns
        -------
        T_0 : float
            The effective temperature at r0.
        """
        # Lipunova, G., Malanchev, K., Shakura, N. (2018). Page 31
        # DOI https://doi.org/10.1007/978-3-319-93009-1_1
        sigma_sb = constants.sigma_sb.cgs.value
        G = constants.G.cgs.value
        return 2 ** (3 / 4) * (3 / 7) ** (7 / 4) * (G * M * Mdot / (np.pi * sigma_sb * r_in**3)) ** (1 / 4)

    def compute_sed(self, times, wavelengths, graph_state, **kwargs):
        """Draw effect-free observations for this object.

        Parameters
        ----------
        times : `numpy.ndarray`
            A length T array of rest frame timestamps.
        wavelengths : `numpy.ndarray`, optional
            A length N array of wavelengths (in angstroms).
        graph_state : `GraphState`
            An object mapping graph parameters to their values.
        **kwargs : `dict`, optional
            Any additional keyword arguments.

        Returns
        -------
        flux_density : `numpy.ndarray`
            A length T x N matrix of SED values (in nJy).
        """
        params = self.get_local_params(graph_state)
        waves_cm = wavelengths * ANGSTROM_TO_CM
        nu = constants.c.cgs.value / waves_cm
        bh_mass_g = params["blackhole_mass_gram"]

        # Compute the parameters for these wavelengths.
        tau_v = self.compute_tau(params["tau_4000aa"], waves_cm)
        sf_inf = self.compute_structure_function_at_inf(params["sf_inf_4000aa"], waves_cm)

        # Compute the average flux of a standard disk model. Use a factor of 2 (two sides
        # of the disk) to get the total flux.
        fnu_average = self.compute_sed_standard_disk(
            Mdot=params["blackhole_accretion_rate"],  # g/s
            nu=nu,  # Hz
            rin_to_rg=6.0,  # rin / rg, no rotation for now
            i=params["inclination_rad"],  # radians
            d=params["distance"] * PARSEC_TO_CM,  # cm
            M=bh_mass_g,  # g
        )

        # Run the damped random walk.
        delta_m = sample_damped_random_walk(
            times,
            tau_v,
            sf_inf,
            params["t0"],
            rng=self._rng,
        )

        flux_density = 10 ** (-0.4 * delta_m) * fnu_average
        return flux_density



import tempfile
from pathlib import Path

import astropy.units as u
import numpy as np
import pandas as pd
import pytest
from astropy.coordinates import Latitude, Longitude, SkyCoord, angular_separation
from lightcurvelynx.astro_utils.detector_footprint import DetectorFootprint
from lightcurvelynx.math_nodes.ra_dec_sampler import (
    ApproximateMOCSampler,
    ObsTableRADECSampler,
    ObsTableUniformRADECSampler,
    UniformRADEC,
)
from lightcurvelynx.math_nodes.single_value_node import SingleVariableNode
from lightcurvelynx.obstable.opsim import OpSim
from lightcurvelynx.utils.io_utils import write_results_as_hats
from mocpy import MOC
from nested_pandas import NestedFrame


def test_uniform_ra_dec():
    """Test that we can generate numbers from a uniform distribution on a sphere."""
    sampler_node = UniformRADEC(seed=100, node_label="sampler")

    # Test we can generate a single value.
    (ra, dec) = sampler_node.generate(num_samples=1)
    assert 0.0 <= ra <= 360.0
    assert -90.0 <= dec <= 90.0

    # Generate many samples.
    num_samples = 20_000
    state = sampler_node.sample_parameters(num_samples=num_samples)

    all_ra = state["sampler"]["ra"]
    assert len(all_ra) == num_samples
    assert np.all(all_ra >= 0.0)
    assert np.all(all_ra <= 360.0)

    all_dec = state["sampler"]["dec"]
    assert len(all_dec) == num_samples
    assert np.all(all_dec >= -90.0)
    assert np.all(all_dec <= 90.0)

    # Compute histograms of RA and dec values.
    ra_bins = np.zeros(36)
    dec_bins = np.zeros(18)
    for idx in range(num_samples):
        ra_bins[int(all_ra[idx] / 10.0)] += 1
        dec_bins[int((all_dec[idx] + 90.0) / 10.0)] += 1

    # Check that all RA bins have approximately equal samples.
    expected_count = num_samples / 36
    for bin_count in ra_bins:
        assert 0.8 <= bin_count / expected_count <= 1.2

    # Check that the dec bins around the poles have less samples
    # than the bins around the equator.
    assert dec_bins[0] < 0.25 * dec_bins[9]
    assert dec_bins[17] < 0.25 * dec_bins[10]

    # Check that we can generate uniform samples in radians.
    sampler_node2 = UniformRADEC(seed=100, node_label="sampler2", use_degrees=False)
    state2 = sampler_node2.sample_parameters(num_samples=num_samples)

    all_ra = state2["sampler2"]["ra"]
    assert len(all_ra) == num_samples
    assert np.all(all_ra >= 0.0)
    assert np.all(all_ra <= 2.0 * np.pi)

    all_dec = state2["sampler2"]["dec"]
    assert len(all_dec) == num_samples
    assert np.all(all_dec >= -np.pi)
    assert np.all(all_dec <= np.pi)


def test_obstable_ra_dec_sampler():
    """Test that we can sample from am OpSim object."""
    values = {
        "observationStartMJD": np.array([0.0, 1.0, 2.0, 3.0, 4.0]),
        "fieldRA": np.array([15.0, 30.0, 15.0, 0.0, 60.0]),
        "fieldDec": np.array([-10.0, -5.0, 0.0, 5.0, 10.0]),
        "zp_nJy": np.ones(5),
    }
    ops_data = OpSim(values)
    assert len(ops_data) == 5

    sampler_node = ObsTableRADECSampler(ops_data, radius=0.0, in_order=True)
    assert sampler_node.radius == 0.0

    # Test we can generate a single value.
    (ra, dec, time) = sampler_node.generate(num_samples=1)
    assert ra == 15.0
    assert dec == -10.0
    assert time == 0.0

    # Test we can generate multiple observations
    (ra, dec, time) = sampler_node.generate(num_samples=2)
    assert np.allclose(ra, [30.0, 15.0])
    assert np.allclose(dec, [-5.0, 0.0])
    assert np.allclose(time, [1.0, 2.0])

    # Do randomized sampling (with no offset).
    sampler_node2 = ObsTableRADECSampler(ops_data, in_order=False, radius=0.0, seed=100, node_label="sampler")
    state = sampler_node2.sample_parameters(num_samples=5000)

    # Check that the samples are uniform and consistent.
    int_times = state["sampler"]["time"].astype(int)
    assert np.allclose(state["sampler"]["ra"], values["fieldRA"][int_times])
    assert np.allclose(state["sampler"]["dec"], values["fieldDec"][int_times])
    assert len(int_times[int_times == 0]) > 750
    assert len(int_times[int_times == 1]) > 750
    assert len(int_times[int_times == 2]) > 750
    assert len(int_times[int_times == 3]) > 750
    assert len(int_times[int_times == 4]) > 750

    # Do randomized sampling with offsets (using the default OpSim radius).
    sampler_node3 = ObsTableRADECSampler(ops_data, in_order=False, seed=100, node_label="sampler")
    state = sampler_node3.sample_parameters(num_samples=5000)

    # Check that the samples are not all the centers (unique values > 500) but are within
    # the sampling radius of the corresponding pointing. We use int(time) to match the index
    # of the center pointing.
    int_times = state["sampler"]["time"].astype(int)
    assert len(np.unique(state["sampler"]["ra"])) > 500
    assert len(np.unique(state["sampler"]["dec"])) > 500
    for idx in range(5):
        dist = angular_separation(
            state["sampler"]["ra"][int_times == idx] * u.deg,
            state["sampler"]["dec"][int_times == idx] * u.deg,
            values["fieldRA"][idx] * u.deg,
            values["fieldDec"][idx] * u.deg,
        )
        assert np.all(dist <= sampler_node3.radius * u.deg)

    # We fail if no radius is provided by the OpSim or the parameters
    ops_data = OpSim(values, radius=None)
    with pytest.raises(ValueError):
        _ = ObsTableRADECSampler(ops_data)

    # But we can use the OpSim's radius if that is provided.
    ops_data = OpSim(values, radius=1.0)
    sampler_node = ObsTableRADECSampler(ops_data)
    assert sampler_node.radius == 1.0


def test_obstable_ra_dec_sampler_extra():
    """Test that we can sample from an ObsTable-like object with extra columns."""
    values = {
        "time": np.array([0.0, 1.0, 2.0, 3.0, 4.0]),
        "ra": np.array([15.0, 30.0, 15.0, 0.0, 60.0]),
        "dec": np.array([-10.0, -5.0, 0.0, 5.0, 10.0]),
        "zp": np.ones(5),
        "extra": np.array([10, 20, 30, 40, 50]),
    }
    ops_df = pd.DataFrame(values)
    sampler_node = ObsTableRADECSampler(ops_df, radius=0.0, extra_cols=["extra"], in_order=True)
    assert sampler_node.radius == 0.0

    # Test we can generate a single value.
    sample = sampler_node.generate(num_samples=1)
    assert len(sample) == 4
    assert sample[0] == 15.0  # ra
    assert sample[1] == -10.0  # dec
    assert sample[2] == 0.0  # time
    assert sample[3] == 10  # extra

    # We can chain on any of the column names.
    single_node = SingleVariableNode("extra", sampler_node.extra, node_label="single")
    state = single_node.sample_parameters(num_samples=2)
    assert np.allclose(state["single.extra"], [20.0, 30.0])


def test_obstable_ra_dec_sampler_from_hats(test_data_dir):
    """Test that we can sample from a HATS catalog on disk."""
    outer_dict = {
        "id": [0, 1, 2],
        "ra": [10.0, 10.1, 10.2],
        "dec": [-10.0, -9.9, -10.1],
        "nobs": [3, 2, 1],
        "z": [0.1, 0.2, 0.3],
    }
    inner_dict = {
        "mjd": [59000, 59001, 59002, 59000, 59001, 59000],
        "flux": [10.0, 12.0, 11.0, 15.0, 14.0, 13.0],
        "fluxerr": [1.0, 1.0, 1.0, 1.5, 1.5, 1.0],
        "filter": ["g", "r", "i", "g", "r", "i"],
    }
    nested_inds = [0, 0, 0, 1, 1, 2]
    results = NestedFrame(data=outer_dict, index=[0, 1, 2])
    nested_1 = pd.DataFrame(data=inner_dict, index=nested_inds)
    results = results.add_nested(nested_1, "lightcurve")

    with tempfile.TemporaryDirectory() as dir_name:
        dir_path = Path(dir_name, "test_hats")
        write_results_as_hats(dir_path, results)

        sampler_node = ObsTableRADECSampler.from_hats(
            dir_path,
            radius=0.0,
            in_order=True,
            node_label="sampler",
            extra_cols=["z"],
        )
        assert sampler_node.radius == 0.0

        states = sampler_node.sample_parameters(num_samples=3)
        assert np.allclose(states["sampler"]["ra"], [10.0, 10.1, 10.2])
        assert np.allclose(states["sampler"]["dec"], [-10.0, -9.9, -10.1])
        assert np.allclose(states["sampler"]["z"], [0.1, 0.2, 0.3])


def test_opsim_uniform_ra_dec_sampler():
    """Test that we can sample uniformly from am OpSim object."""
    # Create an opsim with two points in different hemispheres.
    values = {
        "observationStartMJD": np.array([0.0, 1.0]),
        "fieldRA": np.array([15.0, 195.0]),
        "fieldDec": np.array([75.0, -75.0]),
        "zp_nJy": np.ones(2),
    }
    ops_data = OpSim(values)
    assert len(ops_data) == 2

    # Use a very large radius so we do not reject too many samples.
    sampler_node = ObsTableUniformRADECSampler(ops_data, radius=30.0, seed=100, node_label="sampler")
    assert sampler_node.radius == 30.0

    # Test we can generate a single value.
    ra, dec = sampler_node.generate(num_samples=1)
    assert ops_data.is_observed(ra, dec, radius=30.0)

    # Test we can generate many observations
    num_samples = 10_000
    ra, dec = sampler_node.generate(num_samples=num_samples)
    assert np.all(ops_data.is_observed(ra, dec, radius=30.0))

    # We should sample roughly uniformly from the two regions.
    northern_mask = dec > 0.0
    assert np.sum(northern_mask) > 0.4 * num_samples
    assert np.sum(northern_mask) < 0.6 * num_samples

    northern_center = SkyCoord(ra=15.0, dec=75.0, unit="deg")
    northern_pts = SkyCoord(ra[northern_mask], dec[northern_mask], unit="deg")
    assert np.all(northern_pts.separation(northern_center) <= 30.0 * u.deg)

    southern_center = SkyCoord(ra=195.0, dec=-75.0, unit="deg")
    southern_pts = SkyCoord(ra[~northern_mask], dec[~northern_mask], unit="deg")
    assert np.all(southern_pts.separation(southern_center) <= 30.0 * u.deg)

    # We fail if neither the OpSim or the sampler have a radius
    ops_data = OpSim(values, radius=None)
    with pytest.raises(ValueError):
        _ = ObsTableUniformRADECSampler(ops_data)

    # But we succeed if the OpSim has a radius.
    ops_data = OpSim(values, radius=10.0)
    sampler_node = ObsTableUniformRADECSampler(ops_data)
    assert sampler_node.radius == 10.0


def test_opsim_uniform_ra_dec_sampler_footprint():
    """Test that we can sample uniformly from am OpSim object with a footprint."""
    # Create an opsim with two points in different hemispheres.
    values = {
        "observationStartMJD": np.array([0.0, 1.0]),
        "fieldRA": np.array([15.0, 195.0]),
        "fieldDec": np.array([75.0, -75.0]),
        "zp_nJy": np.ones(2),
    }

    # Detector is a rectangle of 20 x 30 degrees.
    fp = DetectorFootprint.from_sky_rect(width=20, height=30, pixel_scale=0.1)
    ops_data = OpSim(values, detector_footprint=fp)

    # Test we can generate multiple observations
    num_samples = 200
    sampler_node = ObsTableUniformRADECSampler(ops_data, radius=50.0, seed=100, node_label="sampler")
    ra, dec = sampler_node.generate(num_samples=num_samples)

    # We should sample roughly uniformly from the two regions.
    northern_mask = dec > 0.0
    assert np.sum(northern_mask) > 0.4 * num_samples
    assert np.sum(northern_mask) < 0.6 * num_samples

    # All the points should be visible.
    assert np.all(ops_data.is_observed(ra, dec, radius=50.0))


def test_approximate_moc_sampler():
    """Test that we can create and sample from an ApproximateMOCSampler."""
    longitudes = Longitude([15.0, 90.0], unit="deg")
    latitudes = Latitude([-20.0, 20.0], unit="deg")
    moc = MOC.from_cones(
        lon=longitudes,
        lat=latitudes,
        radius=1.0 * u.deg,
        max_depth=12,
        union_strategy="large_cones",
    )
    moc_sampler = ApproximateMOCSampler(moc)

    # Test we can generate a single value.
    ra, dec = moc_sampler.generate(num_samples=1)
    assert moc.contains_skycoords(SkyCoord(ra, dec, unit="deg"))

    # Test we can generate many observations
    num_samples = 10_000
    ra, dec = moc_sampler.generate(num_samples=num_samples)
    assert np.all(moc.contains_skycoords(SkyCoord(ra, dec, unit="deg")))

    # We should sample roughly uniformly from the two regions.
    northern_mask = dec > 0.0
    assert np.sum(northern_mask) > 0.4 * num_samples
    assert np.sum(northern_mask) < 0.6 * num_samples

    assert np.all(ra[northern_mask] > 88.0)
    assert np.all(ra[northern_mask] < 92.0)
    assert np.all(ra[~northern_mask] > 13.0)
    assert np.all(ra[~northern_mask] < 17.0)
    assert np.all(dec[northern_mask] > 18.0)
    assert np.all(dec[northern_mask] < 22.0)
    assert np.all(dec[~northern_mask] > -22.0)
    assert np.all(dec[~northern_mask] < -18.0)


def test_approximate_moc_sampler_from_file():
    """Test that we can create an ApproximateMOCSampler from a MOC file."""
    longitudes = Longitude([15.0, 90.0], unit="deg")
    latitudes = Latitude([-20.0, 20.0], unit="deg")
    moc = MOC.from_cones(
        lon=longitudes,
        lat=latitudes,
        radius=1.0 * u.deg,
        max_depth=8,
        union_strategy="large_cones",
    )

    with tempfile.TemporaryDirectory() as dir_name:
        for fmt in ["fits", "json", "ascii"]:
            file_path = Path(dir_name, f"test_mock.{fmt}")

            # We can't load a non-existent file.
            with pytest.raises(FileNotFoundError):
                _ = ApproximateMOCSampler.from_file(file_path, format=fmt)

            moc.save(file_path, format=fmt)
            moc_sampler = ApproximateMOCSampler.from_file(file_path, format=fmt)

            assert moc_sampler.healpix_list is not None



"""Model that generate the SED or bandflux of a source based on given predefined observer frame
SED curves at given wavelengths.

Note: If you are interested in generating light curves from band-level curves, use
the LightcurveTemplateModel in src/lightcurvelynx/models/lightcurve_template_model.py
instead.
"""

import logging
import warnings
from pathlib import Path

import astropy.units as u
import numpy as np
import yaml
from citation_compass import cite_inline
from scipy.interpolate import RectBivariateSpline
from tqdm import tqdm

from lightcurvelynx.astro_utils.unit_utils import flam_to_fnu
from lightcurvelynx.math_nodes.given_sampler import GivenValueSampler
from lightcurvelynx.models.physical_model import SEDModel

logger = logging.getLogger(__name__)


class SEDTemplate:
    """A class to hold a grid of SED data over time and wavelength, and provide
    interpolation capabilities. The quantities can use whichever units are desired.

    Attributes
    ----------
    wavelengths : np.ndarray
        A length W array of the wavelengths for the SED.
    times : np.ndarray
        A length T array of the times for the SED relative to the reference epoch.
    interp : scipy.interpolate object
        The type of interpolation to use. One of 'linear' or 'spline'.
    period : float or None
        The period of this data, if it is periodic. Default is None.

    Parameters
    ----------
    grid_data : np.ndarray
        A 2D array of shape (N x 3) containing phases, wavelengths, and fluxes.
    sed_data_t0 : float, optional
        The reference epoch of the input SED template. This is the time stamp of the input
        array that will correspond to t0 in the model. Default is 0.0.
    interpolation_type : str, optional
        The type of interpolation to use. One of 'linear' or 'cubic'. Default is 'linear'.
    periodic : bool, optional
        Whether the SED template is periodic. Default is False.
    baseline : np.ndarray or None, optional
        A length W array of baseline SED values for each wavelength. This is only used
        for non-periodic SED templates when they are not active. Default is None.
    **kwargs : dict
        Additional keyword arguments that are ignored.
    """

    def __init__(
        self,
        grid_data,
        *,
        sed_data_t0=0.0,
        interpolation_type="linear",
        periodic=False,
        baseline=None,
        **kwargs,
    ):
        grid_data = np.asarray(grid_data)
        if grid_data.ndim != 2 or grid_data.shape[1] != 3:
            raise ValueError(
                f"grid_data must be a 2D array with shape (N x 3). Got {grid_data.shape} instead."
            )
        self.phases, self.wavelengths, sed_values = SEDTemplate._three_column_to_matrix(grid_data)

        # Apply the sed_data_t0 offset to the phases to get the times.
        self.times = self.phases - sed_data_t0

        # Validate the baseline data.
        if baseline is not None:
            baseline = np.asarray(baseline)
            if baseline.shape != (len(self.wavelengths),):
                raise ValueError(
                    f"baseline shape {baseline.shape} must match wavelengths shape {self.wavelengths.shape}."
                )
        self.baseline = baseline

        # If the SED template is periodic, validate the input data and compute the period.
        if periodic:
            if len(self.times) < 2:
                raise ValueError("At least two time points are required for periodic SED templates.")
            if not np.isclose(self.times[0], 0.0):
                self.times = self.times - self.times[0]
            if not np.allclose(sed_values[0, :], sed_values[-1, :]):
                raise ValueError(
                    "For periodic SED templates, the first and last SED values for "
                    "each wavelength must match."
                )
            self.period = self.times[-1] - self.times[0]
        else:
            self.period = None

        # Set up the interpolation object for this SED.
        interp_degree = 3 if interpolation_type == "cubic" else 1
        self.interp = RectBivariateSpline(
            self.times,
            self.wavelengths,
            sed_values,
            kx=interp_degree,
            ky=interp_degree,
        )

    @classmethod
    def from_file(cls, file_path, **kwargs):
        """Create a SEDTemplate from a file containing three-column data.

        Parameters
        ----------
        file_path : str or Path
            The path to the file containing the SED data.
        **kwargs : dict
            Additional keyword arguments to pass to the SEDTemplate constructor.

        Returns
        -------
        SEDTemplate
            The created SEDTemplate instance.
        """
        file_path = Path(file_path)
        if not file_path.exists():
            raise FileNotFoundError(f"SED data file not found: {file_path}.")

        data = np.loadtxt(file_path, comments="#")
        return cls(data, **kwargs)

    @classmethod
    def from_components(cls, times, wavelengths, sed_values, **kwargs):
        """Create a SEDTemplate from separate time, wavelength, and SED value arrays. This is
        a convenience method that packs the data into (and then later unpacks the data from) the
        required three-column format.

        Parameters
        ----------
        times : np.ndarray
            A length T array of times.
        wavelengths : np.ndarray
            A length W array of wavelengths.
        sed_values : np.ndarray
            A 2D array of shape (T x W) containing the SED values.
        **kwargs : dict
            Additional keyword arguments to pass to the SEDTemplate constructor.
        """
        times = np.asarray(times)
        wavelengths = np.asarray(wavelengths)
        sed_values = np.asarray(sed_values)

        if sed_values.shape != (len(times), len(wavelengths)):
            raise ValueError(
                f"sed_values shape {sed_values.shape} must match (len(times), len(wavelengths)) "
                f"= ({len(times)}, {len(wavelengths)})."
            )

        # Pack the data into three-column format.
        grid_data = np.zeros((len(times) * len(wavelengths), 3))
        row = 0
        for t_idx, time in enumerate(times):
            for w_idx, wavelength in enumerate(wavelengths):
                grid_data[row, 0] = time
                grid_data[row, 1] = wavelength
                grid_data[row, 2] = sed_values[t_idx, w_idx]
                row += 1

        return cls(grid_data, **kwargs)

    def evaluate_sed(self, times, wavelengths):
        """Evaluate the SED at the given times and wavelengths.

        Parameters
        ----------
        times : np.ndarray
            A length T array of times (in the given units) at which to evaluate the SED
            (relative to t0).
        wavelengths : np.ndarray
            A length W array of wavelengths (in the given units) at which to evaluate the SED.

        Returns
        -------
        sed_values : np.ndarray
            A (T x W) matrix of SED values (in the given units) at the given times and wavelengths.
        """
        if self.period is None:
            sed_values = np.zeros((len(times), len(wavelengths)))

            in_range = (times >= self.times[0]) & (times <= self.times[-1])
            sed_values[in_range, :] = self.interp(times[in_range], wavelengths, grid=True)

            if self.baseline is not None:
                sed_values[~in_range, :] = self.baseline[np.newaxis, :]
        else:
            # Create the modulo times for periodic evaluation and an inverse mapping to original order.
            times = np.mod(times, self.period)
            argsort_idx = np.argsort(times)
            inv_idx = np.zeros_like(argsort_idx)
            inv_idx[argsort_idx] = np.arange(len(times))

            sed_values = self.interp(times[argsort_idx], wavelengths, grid=True)
            sed_values = sed_values[inv_idx, :]
        return sed_values

    @staticmethod
    def _three_column_to_matrix(data):
        """Convert 3-column SED data to a matrix form.

        Parameters
        ----------
        data : np.ndarray
            A 2D array of shape (N x 3) containing phases, wavelengths, and fluxes.

        Returns
        -------
        unique_phases : np.ndarray
            A length T array of unique phases.
        unique_wavelengths : np.ndarray
            A length W array of unique wavelengths.
        sed_matrix : np.ndarray
            A 2D array of shape (T x W) with fluxes.
        """
        phases = data[:, 0]
        wavelengths = data[:, 1]
        fluxes = data[:, 2]

        unique_wavelengths = np.unique(wavelengths)
        unique_phases = np.unique(phases)

        sed_matrix = np.zeros((len(unique_phases), len(unique_wavelengths)))
        for wave, phase, flux_val in zip(wavelengths, phases, fluxes, strict=False):
            wave_idx = np.where(unique_wavelengths == wave)[0][0]
            phase_idx = np.where(unique_phases == phase)[0][0]
            sed_matrix[phase_idx, wave_idx] = flux_val

        return unique_phases, unique_wavelengths, sed_matrix


class SEDTemplateModel(SEDModel):
    """A model that generates either the SED or bandflux of a source based on
    SED values at given times and wavelengths.

    SEDTemplateModel supports both periodic and non-periodic data. If the template
    is not periodic then the given values will be interpolated during the time range
    of the template. Values outside the time range (before and after) will be set to
    the baseline value for that wavelength (0.0 by default).

    Parameterized values include:
      * dec - The object's declination in degrees.
      * ra - The object's right ascension in degrees.
      * t0 - The t0 of the zero phase (if applicable), date.

    Notes
    -----
    If you are interested in generating light curves from band-level curves, use
    the LightcurveTemplateModel in src/lightcurvelynx/models/lightcurve_template_model.py
    instead.

    Attributes
    ----------
    template : SEDTemplate
        The data for the SED, including the times and bandfluxes in each filter.

    Parameters
    ----------
    template : numpy.ndarray or SEDTemplate
        The SED template information can be passed as either:
        1) a SEDTemplate instance, or
        2) a numpy array of shape (T, 3) array where the first column is phase (in days), the
        second column is wavelength (in Angstroms), and the third column is the SED value (in nJy).
    sed_data_t0 : float or None, optional
        The reference epoch of the input template. This is the time stamp of the input
        array that will correspond to t0 in the model. This is only required if the template
        is passed as a numpy array. Default is None.
    interpolation_type : str, optional
        The type of interpolation to use. One of 'linear' or 'cubic'. Default is 'linear'.
    periodic : bool, optional
        Whether the template is periodic. Default is False.
    baseline : np.ndarray or None, optional
        A length W array of baseline SED values for each wavelength. This is only used
        for non-periodic templates when they are not active. Default is None.
    """

    def __init__(
        self,
        template,
        *,
        sed_data_t0=None,
        interpolation_type="linear",
        periodic=False,
        baseline=None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        # Store the template data, parsing out different formats if needed.
        if isinstance(template, SEDTemplate):
            self.template = template
        else:
            if sed_data_t0 is None:
                raise ValueError("sed_data_t0 must be provided when template is not a SEDTemplate instance.")
            self.template = SEDTemplate(
                template,
                sed_data_t0=sed_data_t0,
                periodic=periodic,
                baseline=baseline,
                interpolation_type=interpolation_type,
            )

        # Check that t0 is set.
        if "t0" not in kwargs or kwargs["t0"] is None:
            raise ValueError("SED template models require a t0 parameter.")

    @classmethod
    def from_file(cls, file_path, **kwargs):
        """Create a SEDTemplateModel from a file containing three-column data.

        Parameters
        ----------
        file_path : str or Path
            The path to the file containing the SED data.
        **kwargs : dict
            Additional keyword arguments to pass to the SEDTemplateModel constructor.

        Returns
        -------
        SEDTemplateModel
            The created SEDTemplateModel instance.
        """
        template = SEDTemplate.from_file(file_path, **kwargs)
        return cls(template, **kwargs)

    @property
    def times(self):
        """The times of the template data (in days)."""
        return self.template.times

    @property
    def wavelengths(self):
        """The wavelengths of the template data (in Angstroms)."""
        return self.template.wavelengths

    def compute_sed(self, times, wavelengths, graph_state):
        """Draw effect-free observer frame flux densities.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of observer frame timestamps in MJD.
        wavelengths : numpy.ndarray, optional
            A length N array of observer frame wavelengths (in angstroms).
        graph_state : GraphState
            An object mapping graph parameters to their values.

        Returns
        -------
        flux_density : numpy.ndarray
            A length T x N matrix of observer frame SED values (in nJy).
        """
        shifted_times = times - self.get_param(graph_state, "t0")
        return self.template.evaluate_sed(shifted_times, wavelengths)


class MultiSEDTemplateModel(SEDModel):
    """A MultiSEDTemplateModel randomly selects a SED template at each evaluation
    computes the flux from that source at given times and wavelengths.

    MultiSEDTemplateModel supports both periodic and non-periodic templates. See the
    SEDTemplate documentation for details on how each template is handled.

    Parameterized values include:
      * dec - The object's declination in degrees.
      * ra - The object's right ascension in degrees.
      * t0 - The t0 of the zero phase (if applicable), date.

    Attributes
    ----------
    templates : list of SEDTemplate
        The data for the templates, such as the times and bandfluxes in each filter.

    Parameters
    ----------
    templates : list of SEDTemplate
        The data for the templates, such as the times and bandfluxes in each filter.
    weights : numpy.ndarray, optional
        A length N array indicating the relative weight from which to select
        a template at random. If None, all templates will be weighted equally.
    """

    def __init__(
        self,
        templates,
        *,
        weights=None,
        **kwargs,
    ):
        # Validate the input templates.
        for tmp in templates:
            if not isinstance(tmp, SEDTemplate):
                raise TypeError("Each template must be an instance of SEDTemplate.")
        self.templates = templates

        super().__init__(**kwargs)

        all_inds = [i for i in range(len(templates))]
        self._sampler_node = GivenValueSampler(all_inds, weights=weights)
        self.add_parameter(
            "selected_template",
            value=self._sampler_node,
            allow_gradient=False,
            description="Index of the SED template selected for sampling.",
        )

    def __len__(self):
        """Get the number of SED templates."""
        return len(self.templates)

    def compute_sed(self, times, wavelengths, graph_state):
        """Draw effect-free observer frame flux densities.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of observer frame timestamps in MJD.
        wavelengths : numpy.ndarray, optional
            A length N array of observer frame wavelengths (in angstroms).
        graph_state : GraphState
            An object mapping graph parameters to their values.

        Returns
        -------
        flux_density : numpy.ndarray
            A length T x N matrix of observer frame SED values (in nJy). These are generated
            from non-overlapping box-shaped SED basis functions for each filter and
            scaled by the light curve values.
        """
        # Use the light curve selected by the sampler node to compute the flux density.
        model_ind = self.get_param(graph_state, "selected_template")
        shifted_times = times - self.get_param(graph_state, "t0")
        return self.templates[model_ind].evaluate_sed(shifted_times, wavelengths)


class SIMSEDModel(MultiSEDTemplateModel):
    """Generate fluxes from SIMSED-formated data.

    Parameterized values include:
      * dec - The object's declination in degrees.
      * distance - The object's luminosity distance in pc.
      * ra - The object's right ascension in degrees.
      * t0 - The t0 of the zero phase (if applicable), date.

    Attributes
    ----------
    templates : list of SEDTemplate
        The data for the templates, such as the times and bandfluxes in each filter.
    flux_scale : float
        A scale factor to apply to all fluxes read from the SIMSED data files.
    """

    def __init__(self, templates, flux_scale=1.0, **kwargs):
        self.flux_scale = flux_scale
        super().__init__(templates, **kwargs)
        if not self.has_valid_param("distance"):
            raise ValueError(
                "SIMSEDModel requires a valid 'distance' parameter representing luminosity distance in pc. "
                "This can be specified as 'distance' directly or derived by a combination of the 'redshift' "
                "and 'cosmology' parameters."
            )

    @classmethod
    def from_dir(cls, simsed_dir, **kwargs):
        """Read SNANA-formatted data from a directory and create a SIMSEDModel.

        Parameters
        ----------
        simsed_dir : str or Path
            The directory containing the SIMSED-formatted data files.
        **kwargs : dict
            Additional keyword arguments to pass to the SIMSEDModel constructor.

        Returns
        -------
        SIMSEDModel
            The created SIMSEDModel instance.
        """
        simsed_dir = Path(simsed_dir)
        logger.debug(f"Reading SIMSED data from {simsed_dir}.")
        file_names, simsed_params = SIMSEDModel._read_simsed_info_file(simsed_dir)

        # Extract the parameters that we need.
        if "FLUX_SCALE" not in simsed_params:
            warnings.warn("SIMSED SED.INFO file does not contain a FLUX_SCALE parameter. Using 1.0.")
            flux_scale = 1.0
        else:
            flux_scale = float(simsed_params["FLUX_SCALE"])

        # Read in each template file.
        templates = []
        for file_name in tqdm(file_names, total=len(file_names), desc="Loading", unit="file"):
            templates.append(SIMSEDModel._read_simsed_data_file(simsed_dir / file_name))

        # Add a citation for this data file.
        cite_inline(
            "SIMSED Data",
            f"SIMSED data files from {simsed_dir}. Check the SED.INFO file for citation information.",
        )

        return cls(templates, flux_scale=flux_scale, **kwargs)

    @staticmethod
    def _read_simsed_info_file(simsed_dir):
        """Read the SED.INFO file to get the list of template files and their properties.

        Parameters
        ----------
        simsed_dir : Path
            The directory containing the SIMSED-formatted data files.

        Returns
        -------
        file_names : list of Path
            A list of paths to the template files.
        parameters : dict
            A dictionary of parameters read from the SED.INFO file.
        """
        info_file = simsed_dir / "SED.INFO"
        if not info_file.exists():
            raise FileNotFoundError(f"SED.INFO file not found in {simsed_dir}.")
        logger.debug(f"Reading SIMSED data from {info_file}.")

        parameters = {}
        file_names = []
        with open(info_file, "r") as f:
            # Read the header as YAML data.
            parameters = yaml.safe_load(f)

            # Reset the file pointer to read line by line for the file names,
            # because the YAML parser will only return the first one.
            f.seek(0)
            for line in f:
                line = line.strip()
                if line and line.upper().startswith("SED:"):
                    tokens = line.split()
                    file_names.append(simsed_dir / tokens[1].strip())
        return file_names, parameters

    @staticmethod
    def _read_simsed_data_file(file_path):
        """Read a simsed data file to get an individual template.

        Parameters
        ----------
        file_path : Path
            The path to the SIMSED-formatted data file.

        Returns
        -------
        SEDTemplate
            The SED template data extracted from the file.
        """
        if not file_path.exists() and file_path.suffix != ".gz":
            # If the file is not found, check for a .gz version.
            file_path = file_path.with_suffix(file_path.suffix + ".gz")
        if not file_path.exists():
            raise FileNotFoundError(f"SIMSED data file not found: {file_path}.")

        # Read in the data file.
        data = np.loadtxt(file_path, comments="#")
        if data.ndim != 2 or data.shape[1] != 3:
            raise ValueError(
                f"SIMSED data file {file_path} must have three columns (phase, wavelength, flux)."
            )

        sed_data = SEDTemplate(
            data,
            sed_data_t0=0.0,
            interpolation_type="linear",
            periodic=False,
        )
        return sed_data

    def compute_sed(self, times, wavelengths, graph_state):
        """Draw effect-free observer frame flux densities.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of observer frame timestamps in MJD.
        wavelengths : numpy.ndarray, optional
            A length N array of observer frame wavelengths (in angstroms).
        graph_state : GraphState
            An object mapping graph parameters to their values.

        Returns
        -------
        flux_density : numpy.ndarray
            A length T x N matrix of observer frame SED values (in nJy).
        """
        sed_values_flam = super().compute_sed(times, wavelengths, graph_state)

        # SIMSED data files provide flux in units of erg/cm²/s/Å at 10 pc (with a scale factor), so we
        # apply the scale factor, account for the actual distance, and convert it to fnu in nJy.
        luminosity_distance_pc = self.get_param(graph_state, "distance")
        if luminosity_distance_pc is None or luminosity_distance_pc <= 0:
            raise ValueError(
                f"Received invalid luminosity distance (pc) in SIMSED model {luminosity_distance_pc}."
            )
        sed_values_fnu = flam_to_fnu(
            flux_flam=sed_values_flam * self.flux_scale * (10 / luminosity_distance_pc) ** 2,
            wavelengths=wavelengths,
            wave_unit=u.AA,
            flam_unit=u.erg / u.s / u.cm**2 / u.AA,
            fnu_unit=u.nJy,
        )
        return sed_values_fnu



import numpy as np
from lightcurvelynx.astro_utils.coordinate_utils import dedup_coords, ra_dec_to_cartesian


def test_ra_dec_to_cartesian():
    """Test the conversion from RA/Dec to Cartesian coordinates."""
    # Test known values
    ra = np.array([0, 90, 180, 270, 90, 0, 45.0])
    dec = np.array([0, 0, 0, 0, 90, -90, 30.0])
    x, y, z = ra_dec_to_cartesian(ra, dec)

    expected_x = np.array([1, 0, -1, 0, 0, 0, np.sqrt(6) / 4.0])
    expected_y = np.array([0, 1, 0, -1, 0, 0, np.sqrt(6) / 4.0])
    expected_z = np.array([0, 0, 0, 0, 1, -1, 1.0 / 2.0])

    np.testing.assert_allclose(x, expected_x, atol=1e-5)
    np.testing.assert_allclose(y, expected_y, atol=1e-5)
    np.testing.assert_allclose(z, expected_z, atol=1e-5)


def test_dedup_coords():
    """Test the deduplication of coordinates."""
    pts = np.array(
        [
            [0, 0],  # First unique point
            [0.01, 0.01],  # Duplicate of first point
            [-0.01, -0.01],  # Duplicate of first point
            [1, 1],  # Second unique point
            [100, -10],  # Third unique point
            [99.9999, -10.0001],  # Duplicate of third point
            [45, 90],  # Fourth unique point
            [-0.01, -0.01],  # Duplicate of first point
            [1.0001, 1.0001],  # Duplicate of second point
            [100.0001, -9.9999],  # Duplicate of third point
            [270.0, -90.0],  # Fifth unique point
        ]
    )
    unique_ra, unique_dec, unique_indices = dedup_coords(pts[:, 0], pts[:, 1], threshold=0.02)
    expected_inds = np.array([0, 3, 4, 6, 10])
    np.testing.assert_array_equal(unique_indices, expected_inds)
    np.testing.assert_allclose(unique_ra, pts[expected_inds, 0], atol=1e-8)
    np.testing.assert_allclose(unique_dec, pts[expected_inds, 1], atol=1e-8)
